diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/doc/README.hier postgresql-7.3rc1-evg/doc/README.hier
*** postgresql-7.3rc1/doc/README.hier	Thu Jan  1 04:00:00 1970
--- postgresql-7.3rc1-evg/doc/README.hier	Thu Nov 28 20:50:52 2002
***************
*** 0 ****
--- 1,151 ----
+ 
+ WHAT'S THIS
+ 
+ This is a patch which allows PG to make hierarchical queries a la Oracle do.
+ 
+ Copyright (c) Evgen Potemkin 2002, evgent@terminal.ru
+ Distributed under same terms as PostgreSQL itself.
+ 
+ INSTALLATION
+ 
+ tar xzf postgresql-7.3rc1.tar.gz
+ patch -p0 < hier.diff
+ cd postgresql-7.3rc1/
+ ./configure
+ ... so on
+ 
+ SYNOPSIS
+ 	
+ >create table data (id int4, pnt int4, data varchar(20));
+ 
+ data:
+ +----+-----+-------------+
+ | id | pnt |    data     |
+ +----+-----+-------------+
+ | 0  | 0   | root        |
+ | 1  | 0   | (1 leaf l1) |
+ | 2  | 0   | (2 leaf l1) |
+ | 3  | 0   | (3 leaf l1) |
+ | 4  | 1   | (11 leaf l2)|
+ | 5  | 1   | (12 leaf l2)|
+ | 6  | 1   | (13 leaf l2)|
+ | 7  | 6   | (31 leaf l2)|
+ | 8  | 6   | (32 leaf l2)|
+ +----+-----+-------------+
+ 
+ > SELECT * FROM data CONNECT BY id PRIOR pnt START WITH id=0;
+ 
+ output:
+ 
+  id | pnt |    data     | _level_
+ ----+-----+-------------+---------
+  0  | 0   | root        | 1
+  1  | 0   | (1 leaf l1) | 2
+  4  | 1   | (11 leaf l2)| 3
+  5  | 1   | (12 leaf l2)| 3
+  6  | 1   | (13 leaf l2)| 3
+  2  | 0   | (2 leaf l1) | 2
+  3  | 0   | (3 leaf l1) | 2
+  7  | 6   | (31 leaf l2)| 3
+  8  | 6   | (32 leaf l2)| 3
+                         
+ DESCRIPTION             
+                         
+ Hierarchical query.     
+ 
+ Lets tree is like: 
+                           (root)
+ 						      	    /   |    \
+ 			      					/     |      \
+        			(1 leaf l1) (2 leaf l1) (3 leaf l1)
+       			/    |    \                   |     \
+           /      |     \                  |       \
+ (11 leaf l2) (12 leaf l2) (13 leaf l2) (31 leaf l2) (32 leaf l2)
+ 
+ so lets define hierarchical query as query which return us a set like:
+ +------------+
+ |root        |
+ |(1 leaf l1) |
+ |(11 leaf l2)|
+ |(12 leaf l2)|
+ |(13 leaf l2)|
+ |(2 leaf l1) |
+ |(3 leaf l1) |
+ |(31 leaf l2)|
+ |(32 leaf l2)|
+ +------------+
+ 
+ i.e. first is root,then first child of root, 
+ then first child of first child, if any, or second child otherwise, ... so on.
+ 
+ Reason.
+ 
+ Often hierarchical queries are made by adding special field which stores a position
+ of a row in the tree, and then make "SELECT ... ORDER BY 'hier_position_field'" 
+ over the table. disability of such approach is that we can't sort leaves in some 
+ custom order.
+ 
+ This patch is based on internal PG's query processing and havn't such disadvantage.
+ 
+ Syntax.
+ 
+ SELECT ... FROM ... [ WHERE condition ] hier_clause
+   [ HAVING condition [, ...]] [ ORDER BY ... ] [ LIMIT ... ] [ OFFSET ... ]
+ 
+ hier_clause: connect_clause start_with_clause
+ 		| start_with_clause connect_clause
+ 		| /* EMPTY */
+ 		;
+ connect_clause: CONNECT BY PRIOR c_expr qual_Op c_expr
+ 		| CONNECT BY c_expr qual_Op PRIOR c_expr
+ 		;
+ start_with_clause: START WITH a_expr
+ 		;
+ 
+ ... means the same as usual SELECT
+ qual_Op is an operator.
+ c_expr is a column name or function call
+ expr is a WHERE-like expression
+ 
+ WHERE condition
+ 	only rows which pass condition will be used
+ 	
+ CONNECT BY PRIOR c_expr qual_Op c_expr 
+ CONNECT BY c_expr qual_Op PRIOR c_expr 
+ 	expression next to PRIOR is evaluated against parent tuple, another c_expr - against
+ 	child tuple. qual_Op is an operator for comparation, if results of evaluation
+ 	is equal then tuples will be connected.
+ 	qual_Op is not only '=', it may be any comparation operator for types of given expressions.	
+ 	for cases with different parent and child expressions: operator's function is called
+ 		op_func(child_expression, parent_expression).
+ 	for user-defined operators: operator's function must return 0 if values aer equal, 
+ 	  !0 otherwise.
+ 	EX.: CONNECT BY PRIOR id = pnt
+ 	rows will be connected by parent 'id' and child 'pnt' columns.
+ 	i.e. if in some row (A) column 'pnt' is equal to 'id' column of other row (B)
+ 	means that row A is a child of B.
+ 
+ START WITH expr
+ 	rows for which expr is true will be treaten as root rows and other will 
+ 	be connected to them.
+ 
+ HAVING condition
+ 	used to qual rows after building hierarchy. only which satisfy condition will be returned.
+ 
+ 
+ Results.
+ 
+ query return sets like discribed in synopsis. '_level_' is additional column which 
+ added only in case of hierarchical query.
+ it means deep of row in the tree, starts from 1.
+ 
+ Difference with Oracle's hierarchical queries:
+ 	WHERE clause evaluated BEFORE connecting.
+ 	for rows qualification afted they're connected used HAVING.
+ 
+ Misc.	
+ 	Subqueries, views are supported.
+ 	Queries like "select ... from (select * connect by ... start with ...) p,t where p.id = t.id"
+ 	is useless because join (p,t) will resort p and t, and thus destroy hierarchical order.
+ 
+ Have Fun.:)	
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/commands/explain.c postgresql-7.3rc1-evg/src/backend/commands/explain.c
*** postgresql-7.3rc1/src/backend/commands/explain.c	Mon Oct 14 09:26:54 2002
--- postgresql-7.3rc1-evg/src/backend/commands/explain.c	Thu Nov 28 19:15:21 2002
***************
*** 306,311 ****
--- 306,314 ----
  		case T_Hash:
  			pname = "Hash";
  			break;
+  		case T_Conn:
+  			pname = "Hier";
+  			break;
  		default:
  			pname = "???";
  			break;
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/executor/Makefile postgresql-7.3rc1-evg/src/backend/executor/Makefile
*** postgresql-7.3rc1/src/backend/executor/Makefile	Mon May 13 04:43:02 2002
--- postgresql-7.3rc1-evg/src/backend/executor/Makefile	Thu Nov 28 19:15:21 2002
***************
*** 18,24 ****
         nodeHashjoin.o nodeIndexscan.o nodeMaterial.o nodeMergejoin.o \
         nodeNestloop.o nodeFunctionscan.o nodeResult.o nodeSeqscan.o \
         nodeSetOp.o nodeSort.o nodeUnique.o nodeLimit.o nodeGroup.o \
!        nodeSubplan.o nodeSubqueryscan.o nodeTidscan.o spi.o
  
  all: SUBSYS.o
  
--- 18,24 ----
         nodeHashjoin.o nodeIndexscan.o nodeMaterial.o nodeMergejoin.o \
         nodeNestloop.o nodeFunctionscan.o nodeResult.o nodeSeqscan.o \
         nodeSetOp.o nodeSort.o nodeUnique.o nodeLimit.o nodeGroup.o \
!        nodeSubplan.o nodeSubqueryscan.o nodeTidscan.o spi.o nodeConn.o
  
  all: SUBSYS.o
  
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/executor/execAmi.c postgresql-7.3rc1-evg/src/backend/executor/execAmi.c
*** postgresql-7.3rc1/src/backend/executor/execAmi.c	Fri Jun 21 01:29:27 2002
--- postgresql-7.3rc1-evg/src/backend/executor/execAmi.c	Thu Nov 28 19:15:21 2002
***************
*** 37,42 ****
--- 37,43 ----
  #include "executor/nodeSubqueryscan.h"
  #include "executor/nodeFunctionscan.h"
  #include "executor/nodeUnique.h"
+ #include "executor/nodeConn.h"
  
  
  /* ----------------------------------------------------------------
***************
*** 157,162 ****
--- 158,167 ----
  			ExecReScanAppend((Append *) node, exprCtxt, parent);
  			break;
  
+  		case T_Conn:
+  			ExecReScanConn((Conn *) node, exprCtxt, parent);
+  			break;
+ 
  		default:
  			elog(ERROR, "ExecReScan: node type %d not supported",
  				 nodeTag(node));
***************
*** 208,213 ****
--- 213,222 ----
  			ExecTidMarkPos((TidScan *) node);
  			break;
  
+  		case T_Conn:
+  			ExecConnMarkPos((Conn *) node);
+  			break;
+  
  		default:
  			/* don't make hard error unless caller asks to restore... */
  			elog(LOG, "ExecMarkPos: node type %d not supported",
***************
*** 248,253 ****
--- 257,266 ----
  
  		case T_Sort:
  			ExecSortRestrPos((Sort *) node);
+  			break;
+  
+  		case T_Conn:
+  			ExecConnRestrPos((Conn *) node);
  			break;
  
  		default:
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/executor/execProcnode.c postgresql-7.3rc1-evg/src/backend/executor/execProcnode.c
*** postgresql-7.3rc1/src/backend/executor/execProcnode.c	Fri Jun 21 01:29:27 2002
--- postgresql-7.3rc1-evg/src/backend/executor/execProcnode.c	Thu Nov 28 19:15:21 2002
***************
*** 100,105 ****
--- 100,106 ----
  #include "executor/nodeUnique.h"
  #include "miscadmin.h"
  #include "tcop/tcopprot.h"
+ #include "executor/nodeConn.h"
  
  /* ------------------------------------------------------------------------
   *		ExecInitNode
***************
*** 224,229 ****
--- 225,234 ----
  			result = ExecInitAgg((Agg *) node, estate, parent);
  			break;
  
+  		case T_Conn:
+  			result = ExecInitConn((Conn *) node, estate, parent);
+  			break;
+  
  		default:
  			elog(ERROR, "ExecInitNode: node type %d unsupported",
  				 (int) nodeTag(node));
***************
*** 357,362 ****
--- 362,371 ----
  			result = ExecAgg((Agg *) node);
  			break;
  
+  		case T_Conn:
+  			result = ExecConn((Conn *) node);
+  			break;
+  
  		default:
  			elog(ERROR, "ExecProcNode: node type %d unsupported",
  				 (int) nodeTag(node));
***************
*** 444,449 ****
--- 453,461 ----
  		case T_Agg:
  			return ExecCountSlotsAgg((Agg *) node);
  
+  		case T_Conn:
+  			return ExecCountSlotsConn((Conn *) node);
+  
  		default:
  			elog(ERROR, "ExecCountSlotsNode: node type %d unsupported",
  				 (int) nodeTag(node));
***************
*** 570,575 ****
--- 582,591 ----
  			ExecEndAgg((Agg *) node);
  			break;
  
+  		case T_Conn:
+  			ExecEndConn((Conn *) node);
+  			break;
+  
  		default:
  			elog(ERROR, "ExecEndNode: node type %d unsupported",
  				 (int) nodeTag(node));
***************
*** 745,750 ****
--- 761,774 ----
  			}
  			break;
  
+  		case T_Conn:
+  			{
+  				CommonScanState *scanstate = ((Conn *) node)->connstate;
+   
+   				slot = scanstate->cstate.cs_ResultTupleSlot;
+  			}
+ 			break;
+ 			
  		default:
  
  			/*
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/executor/execQual.c postgresql-7.3rc1-evg/src/backend/executor/execQual.c
*** postgresql-7.3rc1/src/backend/executor/execQual.c	Thu Sep  5 01:31:17 2002
--- postgresql-7.3rc1-evg/src/backend/executor/execQual.c	Thu Nov 28 19:15:21 2002
***************
*** 52,57 ****
--- 52,58 ----
  static Datum ExecEvalArrayRef(ArrayRef *arrayRef, ExprContext *econtext,
  				 bool *isNull, ExprDoneCond *isDone);
  static Datum ExecEvalVar(Var *variable, ExprContext *econtext, bool *isNull);
+ static Datum ExecEvalFakeVar(FakeVar *variable, ExprContext *econtext, bool *isNull);
  static Datum ExecEvalOper(Expr *opClause, ExprContext *econtext,
  			 bool *isNull, ExprDoneCond *isDone);
  static Datum ExecEvalDistinct(Expr *opClause, ExprContext *econtext,
***************
*** 383,388 ****
--- 384,442 ----
  	return result;
  }
  
+ /* Same as ExecEvalVar but returns only value from current tuple, or
+  * (Datum) ) if tuple slot if undefined.
+  */
+ static Datum
+ ExecEvalFakeVar(FakeVar *variable, ExprContext *econtext, bool *isNull)
+ {
+ 	Datum		result;
+ 	TupleTableSlot *slot;
+ 	AttrNumber	attnum;
+ 	HeapTuple	heapTuple;
+ 	TupleDesc	tuple_type;
+ 	/*
+ 	 * get the slot we want
+ 	 */
+ 	/* FakeVars can't be OUTER or INNER so get the tuple only from 
+ 	 * the relation being scanned.
+ 	 */
+ 	slot = econtext->ecxt_scantuple;
+ 
+ 	/* if slot is undefined, then we're on layer under Hier where FakeVar not really exist,
+ 	 * so show our fake nature, and return 0.
+ 	 */
+ 	if(!slot){
+ 		if(variable->vartype != INT4OID)
+ 			elog(ERROR,"ExecEvalFakeVar: FakeVars with type other than INT4 not supported (internal error)");
+ 		return Int32GetDatum(0);
+ 	}
+ 	/*
+ 	 * extract tuple information from the slot
+ 	 */
+ 	heapTuple = slot->val;
+ 	tuple_type = slot->ttc_tupleDescriptor;
+ 
+ 	attnum = variable->varattno;
+ 
+ 	/* (See prolog of ExecEvalVar for explanation of this Assert) */
+ 	Assert(attnum <= 0 ||
+ 		   (attnum - 1 <= tuple_type->natts - 1 &&
+ 			tuple_type->attrs[attnum - 1] != NULL &&
+ 		  variable->vartype == tuple_type->attrs[attnum - 1]->atttypid));
+ 
+ 	if (attnum == InvalidAttrNumber)
+ 		elog(ERROR,"ExecEvalFakeVar: invalid attnum %u (internal error)",attnum);
+ 
+ 	result = heap_getattr(heapTuple,	/* tuple containing attribute */
+ 						  attnum,		/* attribute number of desired
+ 										 * attribute */
+ 						  tuple_type,	/* tuple descriptor of tuple */
+ 						  isNull);		/* return: is attribute null? */
+ 
+ 	return result;
+ }
+ 
  /* ----------------------------------------------------------------
   *		ExecEvalParam
   *
***************
*** 1680,1685 ****
--- 1734,1742 ----
  	{
  		case T_Var:
  			retDatum = ExecEvalVar((Var *) expression, econtext, isNull);
+ 			break;
+ 		case T_FakeVar:
+ 			retDatum = ExecEvalFakeVar((FakeVar *) expression, econtext, isNull);
  			break;
  		case T_Const:
  			{
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/executor/nodeConn.c postgresql-7.3rc1-evg/src/backend/executor/nodeConn.c
*** postgresql-7.3rc1/src/backend/executor/nodeConn.c	Thu Jan  1 04:00:00 1970
--- postgresql-7.3rc1-evg/src/backend/executor/nodeConn.c	Thu Nov 28 19:15:21 2002
***************
*** 0 ****
--- 1,344 ----
+ /*------------------------------------------------------------------------- *
+  * nodeConn.c
+  *	  Routines to handle connecting of relations.
+  *
+  * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+  * Portions Copyright (c) 1994, Regents of the University of California
+  *
+  * Based on nodeSort.c, by Evgen Potemkin evgent@terminal.ru, 11.2002
+  *
+  *-------------------------------------------------------------------------
+  */
+ 
+ #include "postgres.h"
+ 
+ #include "executor/execdebug.h"
+ #include "executor/nodeConn.h"
+ #include "utils/tupleconn.h"
+ #include "miscadmin.h"
+ #include "utils/memutils.h"
+ 
+ #include "nodes/print.h"
+ 
+ /* ----------------------------------------------------------------
+  *		ExecConn
+  *
+  *		Connects tuples from the outer subtree of the node using tupleconn,
+  *		which saves the results in a temporary file or memory. After the
+  *		initial call, returns a tuple from the file with each call.
+  *
+  *		Conditions:
+  *		  -- none.
+  *
+  *		Initial States:
+  *		  -- the outer child is prepared to return the first tuple.
+  * ----------------------------------------------------------------
+  */
+ TupleTableSlot *
+ ExecConn(Conn *node)
+ {
+ 	EState	   *estate;
+ 	TupleTableSlot *slot;
+ 	HeapTuple	heapTuple;
+   Plan       *outerNode;
+ 	ConnectState  *connstate;
+ 	ScanDirection dir;
+ 	Tupleconnstate *tupleconnstate;
+ 	bool		should_free;
+ 	List		*squal,*chqual;
+ 	ExprContext *econtext;
+ 	int			head;
+ 	bool		ok;
+ 
+ 	/*
+ 	 * get state info from node
+ 	 */
+ 
+ 	estate = node->plan.state;
+ 	dir = estate->es_direction;
+   connstate=node->connstate;
+ 	tupleconnstate = (Tupleconnstate *) connstate->tupleconnstate;
+ 	
+ 	if (!connstate->conn_Done)
+ 	{
+ 		TupleDesc	tupDesc;
+ 
+ 		SO1_printf("ExecConn: %s\n",
+ 				   "connecting subplan");
+ 
+ 		/*
+ 		 * Want to scan subplan in the forward direction while creating
+ 		 * the data for processing.
+ 		 */
+ 
+ 		/*
+ 		 * Initialize tupleconn module.
+ 		 */
+ 		SO1_printf("ExecConn: %s\n",
+ 				   "calling tupleconn_begin");
+ 
+ 		outerNode = outerPlan((Plan *) node);
+ 		tupDesc = ExecGetTupType(outerNode);
+ 
+     tupleconnstate = tupleconn_begin_heap(SortMem,tupDesc,node->connOp);
+ 		connstate->tupleconnstate = (void *) tupleconnstate;
+ 
+ 
+ 		/*
+ 		 * Scan the subplan and feed all the tuples to tupleconn.
+ 		 */
+     squal=node->startQual;
+ 
+ 		econtext = connstate->csstate.cstate.cs_ExprContext;
+ 
+ 		ResetExprContext(econtext);
+ 
+ 		for (;;)
+ 		{
+ 			slot = ExecProcNode(outerNode, (Plan *) node);
+ 
+ 			if (TupIsNull(slot))
+ 				break;
+ 	    econtext->ecxt_scantuple = slot;
+ 			/* check if tuple is a head of tree */
+ 			head = ExecQual(squal, econtext, false);
+ 			/* store it */
+ 			tupleconn_puttuple(tupleconnstate, (void *) slot->val,head);
+ 		}
+ 
+ 		ExecAssignResultType(&connstate->csstate.cstate, tupDesc, false);
+ 
+     tupleconn_donestoring(tupleconnstate);
+ 		/* connect them.
+ 		 * parentExpr, childExpr, econtext is not needed in tupleconn,
+ 		 * so pass them only for connecting stage.
+ 		 */
+ 		tupleconn_performconn(tupleconnstate,node->parentExpr,node->childExpr,econtext);
+ 		/*
+ 		 * finally set the connected flag to true
+ 		 */
+ 		connstate->conn_Done = true;
+ 		SO1_printf(stderr, "ExecConn: connecting done.\n");
+ 	}
+ 	SO1_printf("ExecConn: %s\n",
+ 			   "retrieving tuple from tupleconn");
+ 
+ 	/*
+ 	 * Get the first or next tuple from tupleconn. Returns NULL if no more
+ 	 * tuples.
+ 	 */
+ 	squal = ((Plan *)node)->qual;
+ 	ok=0;	
+ 	econtext = connstate->csstate.cstate.cs_ExprContext;
+ 	/* everywhere here slot is the same pointer, so get it here */
+ 	slot = connstate->csstate.cstate.cs_ResultTupleSlot;
+ 	/* Apply HAVING clause to post-qual tuples */
+ 	do{
+ 		/* if qual present, clear context */
+ 		if(squal) ResetExprContext(econtext);
+ 		heapTuple = tupleconn_getheaptuple(tupleconnstate,
+ 									   &should_free);
+ 		ExecStoreTuple(heapTuple, slot, InvalidBuffer, should_free);
+ 		if(!squal) break; /* don't qual */
+ 		if(!heapTuple) break; /* no more tuples */
+ 		econtext->ecxt_scantuple = slot;
+ 		/* if tuple satisfy qual, pass it, else get next */
+ 		ok=ExecQual(squal, econtext, false);
+ 	}while(!ok);
+ 	
+ 	if(squal) ResetExprContext(econtext);
+ 	return slot;
+ }
+ 
+ /* Everything below is same as nodeSort.c
+  */
+ /* ----------------------------------------------------------------
+  *		ExecInitConn
+  *
+  *		Creates the run-time state information for the conn node
+  *		produced by the planner and initailizes its outer subtree.
+  * ----------------------------------------------------------------
+  */
+ bool
+ ExecInitConn(Conn *node, EState *estate, Plan *parent)
+ {
+ 	ConnectState *connstate;
+ 	Plan	   *outerPlan;
+ 	
+ 	SO1_printf("ExecInitConn: %s\n",
+ 			   "initializing conn node");
+ 	/*
+ 	 * assign the node's execution state
+ 	 */
+ 	node->plan.state = estate;
+ 
+ 	/*
+ 	 * create state structure
+ 	 */
+ 	connstate = makeNode(ConnectState);
+ 	connstate->conn_Done = false;
+ 	connstate->tupleconnstate = NULL;
+ 
+ 	node->connstate=connstate;
+ 	/*
+ 	 * Miscellaneous initialization
+ 	 */
+   ExecAssignExprContext(estate, &connstate->csstate.cstate); 
+ #define CONN_NSLOTS 2
+ 
+ 	/*
+ 	 * tuple table initialization
+ 	 */
+ 	ExecInitResultTupleSlot(estate, &connstate->csstate.cstate);
+ 	ExecInitScanTupleSlot(estate, &connstate->csstate);
+ 	/*
+ 	 * initializes child nodes
+ 	 */
+ 	outerPlan = outerPlan((Plan *) node);
+ 	ExecInitNode(outerPlan, estate, (Plan *) node);
+ 
+ 	/*
+ 	 * initialize tuple type.  no need to initialize projection info
+ 	 * because this node doesn't do projections.
+ 	 */
+ 	ExecAssignResultTypeFromOuterPlan((Plan *) node, &connstate->csstate.cstate);
+ 	ExecAssignScanTypeFromOuterPlan((Plan *) node, &connstate->csstate);
+ 	connstate->csstate.cstate.cs_ProjInfo = NULL;
+ 
+ 	SO1_printf("ExecInitConn: %s\n",
+ 			   "conn node initialized");
+ 
+ 	return TRUE;
+ }
+ 
+ int
+ ExecCountSlotsConn(Conn *node)
+ {
+ 	return ExecCountSlotsNode(outerPlan((Plan *) node)) +
+ 		ExecCountSlotsNode(innerPlan((Plan *) node))+CONN_NSLOTS;
+ }
+ 
+ /* ----------------------------------------------------------------
+  *		ExecEndConn(node)
+  * ----------------------------------------------------------------
+  */
+ void
+ ExecEndConn(Conn *node)
+ {
+ 	Plan	   *outerPlan;
+ 	ConnectState *connstate;
+ 
+ 	/*
+ 	 * get info from the conn state
+ 	 */
+ 	SO1_printf("ExecEndConn: %s\n",
+ 			   "shutting down conn node");
+   connstate=node->connstate;
+ 
+ 	/*
+ 	 * shut down the subplan
+ 	 */
+ 	outerPlan = outerPlan((Plan *) node);
+ 	ExecEndNode(outerPlan, (Plan *) node);
+ 
+ 	/*
+ 	 * clean out the tuple table
+ 	 */
+ 	ExecClearTuple(connstate->csstate.cstate.cs_ResultTupleSlot);
+ 	ExecClearTuple(connstate->csstate.css_ScanTupleSlot);
+ 
+ 	/*
+ 	 * Release tupleconn resources
+ 	 */
+ 	if (connstate->tupleconnstate != NULL)
+ 		tupleconn_end((Tupleconnstate *) connstate->tupleconnstate);
+ 	connstate->tupleconnstate = NULL;
+ 
+ 	pfree(connstate);
+ 	node->connstate = NULL;
+ 
+ 	SO1_printf("ExecEndConn: %s\n",
+ 			   "conn node shutdown");
+ }
+ 
+ /* ----------------------------------------------------------------
+  *		ExecConnMarkPos
+  *
+  *		Calls tupleconn to save the current position in the processed file.
+  * ----------------------------------------------------------------
+  */
+ void
+ ExecConnMarkPos(Conn *node)
+ {
+ 	ConnectState  *connstate = node->connstate;
+ 
+ 	/*
+ 	 * if we haven't processed yet, just return
+ 	 */
+ 	if (!connstate->conn_Done)
+ 		return;
+ 
+ 	tupleconn_markpos((Tupleconnstate *) connstate->tupleconnstate);
+ }
+ 
+ /* ----------------------------------------------------------------
+  *		ExecConnRestrPos
+  *
+  *		Calls tupleconn to restore the last saved processed file position.
+  * ----------------------------------------------------------------
+  */
+ void
+ ExecConnRestrPos(Conn *node)
+ {
+ 	ConnectState  *connstate = node->connstate;
+ 
+ 	/*
+ 	 * if we haven't processed yet, just return.
+ 	 */
+ 	if (!connstate->conn_Done)
+ 		return;
+ 
+ 	/*
+ 	 * restore the scan to the previously marked position
+ 	 */
+ 	tupleconn_restorepos((Tupleconnstate *) connstate->tupleconnstate);
+ }
+ 
+ void
+ ExecReScanConn(Conn *node, ExprContext *exprCtxt, Plan *parent)
+ {
+ 	ConnectState  *connstate = node->connstate;
+ 
+ 	/*
+ 	 * If we haven't processed yet, just return. If outerplan' chgParam is
+ 	 * not NULL then it will be re-scanned by ExecProcNode, else - no
+ 	 * reason to re-scan it at all.
+ 	 */
+ 	if (!connstate->conn_Done)
+ 		return;
+ 
+ 	ExecClearTuple(node->connstate->csstate.cstate.cs_ResultTupleSlot);
+ 
+ 	/*
+ 	 * If subnode is to be rescanned then we forget previous process results;
+ 	 * we have to re-read the subplan and re-connect.
+ 	 *
+ 	 * Otherwise we can just rewind and rescan the processed output.
+ 	 */
+ 
+ 	if (((Plan *) node)->lefttree->chgParam != NULL)
+ 	{
+ 		connstate->conn_Done = false;
+ 		tupleconn_end((Tupleconnstate *) connstate->tupleconnstate);
+ 		connstate->tupleconnstate = NULL;
+ 	}
+ 	else
+ 		tupleconn_rescan((Tupleconnstate *) connstate->tupleconnstate);
+ 
+ 	/*
+ 	 * if chgParam of subnode is not null then plan will be re-scanned by
+ 	 * first ExecProcNode.
+ 	 */
+ 	if (((Plan *) node)->lefttree->chgParam == NULL)
+ 		ExecReScan(((Plan *) node)->lefttree, exprCtxt, (Plan *) node);
+ }
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/nodes/copyfuncs.c postgresql-7.3rc1-evg/src/backend/nodes/copyfuncs.c
*** postgresql-7.3rc1/src/backend/nodes/copyfuncs.c	Tue Oct 15 03:14:34 2002
--- postgresql-7.3rc1-evg/src/backend/nodes/copyfuncs.c	Thu Nov 28 19:15:21 2002
***************
*** 465,470 ****
--- 465,493 ----
  	return newnode;
  }
  
+ /* ----------------
+  *		_copyConn
+  * ----------------
+  */
+ static Conn *
+ _copyConn(Conn *from)
+ {
+ 	Conn	   *newnode = makeNode(Conn);
+ 	/*
+  	 * copy node superclass fields
+  	 */
+  	CopyPlanFields((Plan *) from, (Plan *) newnode);
+  	
+  	Node_Copy(from,newnode,startQual);
+ 
+ 	Node_Copy(from,newnode,parentExpr);
+  	Node_Copy(from,newnode,childExpr);
+  	
+  	newnode->connOp = newnode->connOp;
+  
+  	return newnode;
+ }
+  
  
  /* ----------------
   *		_copySort
***************
*** 751,756 ****
--- 774,798 ----
  }
  
  /* ----------------
+  *		_copyVar
+  * ----------------
+  */
+ static FakeVar *
+ _copyFakeVar(FakeVar *from)
+ {
+ 	FakeVar		   *newnode = makeNode(FakeVar);
+ 
+ 	/*
+ 	 * copy remainder of node
+ 	 */
+ 	newnode->varattno = from->varattno;
+ 	newnode->vartype = from->vartype;
+ 	newnode->vartypmod = from->vartypmod;
+ 
+ 	return newnode;
+ }
+ 
+ /* ----------------
   *		_copyOper
   * ----------------
   */
***************
*** 1493,1498 ****
--- 1535,1556 ----
  	return newnode;
  }
  
+ /* see _equalHierClause */
+ static HierClause *
+ _copyHierClause(HierClause *from)
+ {
+  	HierClause *newnode = makeNode(HierClause);
+  
+  	Node_Copy(from,newnode,startQual);
+  	
+  	Node_Copy(from,newnode,parentExpr);
+ 	Node_Copy(from,newnode,childExpr);
+ 	
+  	newnode->connOp = from->connOp;
+  
+  	return newnode;
+ }
+  
  static SortClause *
  _copySortClause(SortClause *from)
  {
***************
*** 1802,1807 ****
--- 1860,1867 ----
  	Node_Copy(from, newnode, limitOffset);
  	Node_Copy(from, newnode, limitCount);
  
+  	Node_Copy(from, newnode, hierClause);
+  
  	Node_Copy(from, newnode, setOperations);
  
  	newnode->resultRelations = listCopy(from->resultRelations);
***************
*** 2807,2812 ****
--- 2867,2878 ----
  		case T_SubPlan:
  			retval = _copySubPlan(from);
  			break;
+  		case T_Conn:
+  			retval = _copyConn(from);
+  			break;
+  		case T_HierClause:
+  			retval = _copyHierClause(from);
+  			break;
  
  			/*
  			 * PRIMITIVE NODES
***************
*** 2819,2824 ****
--- 2885,2893 ----
  			break;
  		case T_Expr:
  			retval = _copyExpr(from);
+ 			break;
+ 		case T_FakeVar:
+ 			retval = _copyFakeVar(from);
  			break;
  		case T_Var:
  			retval = _copyVar(from);
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/nodes/equalfuncs.c postgresql-7.3rc1-evg/src/backend/nodes/equalfuncs.c
*** postgresql-7.3rc1/src/backend/nodes/equalfuncs.c	Tue Oct 15 03:14:34 2002
--- postgresql-7.3rc1-evg/src/backend/nodes/equalfuncs.c	Thu Nov 28 19:15:21 2002
***************
*** 124,129 ****
--- 124,142 ----
  }
  
  static bool
+ _equalFakeVar(FakeVar *a, FakeVar *b)
+ {
+ 	if (a->varattno != b->varattno)
+ 		return false;
+ 	if (a->vartype != b->vartype)
+ 		return false;
+ 	if (a->vartypmod != b->vartypmod)
+ 		return false;
+ 
+ 	return true;
+ }
+ 
+ static bool
  _equalOper(Oper *a, Oper *b)
  {
  	if (a->opno != b->opno)
***************
*** 609,614 ****
--- 622,629 ----
  		return false;
  	if (!equal(a->limitCount, b->limitCount))
  		return false;
+ 	if (!equal(a->hierClause, b->hierClause))
+ 		return false;
  	if (!equal(a->setOperations, b->setOperations))
  		return false;
  	if (!equali(a->resultRelations, b->resultRelations))
***************
*** 624,629 ****
--- 639,663 ----
  }
  
  static bool
+ _equalHierClause(HierClause *a, HierClause *b)
+ {
+ 	if (!equal(a->startQual, b->startQual))
+ 		return false;
+ 	if (!equal(a->parentExpr, b->parentExpr))
+ 		return false;
+ 	if (!equal(a->childExpr, b->childExpr))
+ 		return false;
+ 	/* we can skip comparing the connOpName because in the DB
+ 	 * can't exist at the same time several operators with
+ 	 * same name and parameters, it's checked by PG.
+ 	 */
+ 	if (a->connOp != b->connOp)
+ 		return false;
+ 
+ 	return true;
+ }
+ 
+ static bool
  _equalInsertStmt(InsertStmt *a, InsertStmt *b)
  {
  	if (!equal(a->relation, b->relation))
***************
*** 2033,2038 ****
--- 2067,2075 ----
  		case T_Var:
  			retval = _equalVar(a, b);
  			break;
+ 		case T_FakeVar:
+ 			retval = _equalFakeVar(a, b);
+ 			break;
  		case T_Oper:
  			retval = _equalOper(a, b);
  			break;
***************
*** 2432,2437 ****
--- 2469,2477 ----
  			break;
  		case T_InsertDefault:
  			retval = _equalInsertDefault(a, b);
+ 			break;
+ 		case T_HierClause:
+ 			retval = _equalHierClause(a, b);
  			break;
  
  		default:
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/nodes/outfuncs.c postgresql-7.3rc1-evg/src/backend/nodes/outfuncs.c
*** postgresql-7.3rc1/src/backend/nodes/outfuncs.c	Tue Oct 15 03:14:34 2002
--- postgresql-7.3rc1-evg/src/backend/nodes/outfuncs.c	Thu Nov 28 19:15:21 2002
***************
*** 295,300 ****
--- 295,303 ----
  	appendStringInfo(str, " :limitCount ");
  	_outNode(str, node->limitCount);
  
+  	appendStringInfo(str, " :hierClause ");
+  	_outNode(str, node->hierClause);
+   
  	appendStringInfo(str, " :setOperations ");
  	_outNode(str, node->setOperations);
  
***************
*** 305,310 ****
--- 308,328 ----
  }
  
  static void
+ _outHierClause(StringInfo str, HierClause *node)
+ {
+  	appendStringInfo(str, " HIERCLAUSE :startQual ");
+  	_outNode(str, node->startQual);
+ 	
+  	appendStringInfo(str, " :parentExpr ");
+  	_outNode(str, node->parentExpr);
+  	
+  	appendStringInfo(str, " :childExpr ");
+  	_outNode(str, node->childExpr);
+  	
+  	appendStringInfo(str, " :connOp %u ",node->connOp);
+ }
+  
+ static void
  _outSortClause(StringInfo str, SortClause *node)
  {
  	appendStringInfo(str, " SORTCLAUSE :tleSortGroupRef %u :sortop %u ",
***************
*** 592,597 ****
--- 610,633 ----
  	appendStringInfo(str, " :keycount %d ", node->keycount);
  }
  
+ /*
+  *	Conn is a subclass of Plan
+  */
+ static void
+ _outConn(StringInfo str, Conn *node)
+ {
+  	appendStringInfo(str, " CONN ");
+  	_outPlanInfo(str, (Plan *) node);	
+  	appendStringInfo(str, " :startQual ");
+  	_outNode(str, node->startQual);
+ 
+  	appendStringInfo(str, " :parentExpr ");
+  	_outNode(str, node->parentExpr);
+  	appendStringInfo(str, " :childExpr ");
+  	_outNode(str, node->childExpr);
+  	appendStringInfo(str, " :connOp %u ",node->connOp);
+ }
+  
  static void
  _outAgg(StringInfo str, Agg *node)
  {
***************
*** 768,780 ****
  					 node->vartype,
  					 node->vartypmod);
  
! 	appendStringInfo(str, " :varlevelsup %u :varnoold %u :varoattno %d",
  					 node->varlevelsup,
  					 node->varnoold,
  					 node->varoattno);
  }
  
  /*
   *	Const is a subclass of Expr
   */
  static void
--- 804,829 ----
  					 node->vartype,
  					 node->vartypmod);
  
! 	appendStringInfo(str, " :varlevelsup %u :varnoold %u :varoattno %d ",
  					 node->varlevelsup,
  					 node->varnoold,
  					 node->varoattno);
  }
  
  /*
+  *	FakeVar is a subclass of Expr
+  */
+ static void
+ _outFakeVar(StringInfo str, FakeVar *node)
+ {
+ 	appendStringInfo(str,
+ 				" FAKEVAR :varattno %d :vartype %u :vartypmod %d ",
+ 					 node->varattno,
+ 					 node->vartype,
+ 					 node->vartypmod);
+ }
+ 
+ /*
   *	Const is a subclass of Expr
   */
  static void
***************
*** 1660,1665 ****
--- 1709,1717 ----
  			case T_Var:
  				_outVar(str, obj);
  				break;
+ 			case T_FakeVar:
+ 				_outFakeVar(str, obj);
+ 				break;
  			case T_Const:
  				_outConst(str, obj);
  				break;
***************
*** 1777,1782 ****
--- 1829,1840 ----
  			case T_FuncCall:
  				_outFuncCall(str, obj);
  				break;
+  			case T_Conn:
+  				_outConn(str, obj);
+  				break;
+  			case T_HierClause:
+  				_outHierClause(str, obj);
+  				break;
  
  			default:
  				elog(WARNING, "_outNode: don't know how to print type %d ",
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/nodes/readfuncs.c postgresql-7.3rc1-evg/src/backend/nodes/readfuncs.c
*** postgresql-7.3rc1/src/backend/nodes/readfuncs.c	Tue Oct 15 03:14:34 2002
--- postgresql-7.3rc1-evg/src/backend/nodes/readfuncs.c	Thu Nov 28 19:15:21 2002
***************
*** 183,188 ****
--- 183,191 ----
  	token = pg_strtok(&length); /* skip :limitCount */
  	local_node->limitCount = nodeRead(true);
  
+  	token = pg_strtok(&length); /* skip :hierClause */
+  	local_node->hierClause = nodeRead(true);
+  
  	token = pg_strtok(&length); /* skip :setOperations */
  	local_node->setOperations = nodeRead(true);
  
***************
*** 195,200 ****
--- 198,232 ----
  }
  
  /* ----------------
+  *		_readHierClause
+  * ----------------
+  */
+ static HierClause *
+ _readHierClause(void)
+ {
+  	HierClause *local_node;
+  	char	   *token;
+  	int			length;
+  
+  	local_node = makeNode(HierClause);
+  
+  	token = pg_strtok(&length); /* skip :startQual */
+  	local_node->startQual = (Node *) nodeRead(true); /* get startQual */
+  
+  	token = pg_strtok(&length); /* skip :parentExpr */
+  	local_node->parentExpr = (Node *) nodeRead(true); /* get startQual */
+  
+  	token = pg_strtok(&length); /* skip :childExpr */
+  	local_node->childExpr = (Node *) nodeRead(true); /* get startQual */
+  
+  	token = pg_strtok(&length); /* skip :connOp */
+  	token = pg_strtok(&length); /* get connOp */
+  	local_node->connOp = atooid(token);
+  
+  	return local_node;
+ }
+  
+ /* ----------------
   *		_readNotifyStmt
   * ----------------
   */
***************
*** 673,678 ****
--- 705,744 ----
  }
  
  /* ----------------
+  *		_readConn
+  *
+  *	Conn is a subclass of Plan
+  * ----------------
+  */
+ static Conn *
+ _readConn(void)
+ {
+  	Conn	   *local_node;
+  	char	   *token;
+  	int			length;
+  
+  	local_node = makeNode(Conn);
+  
+  	_getPlan((Plan *) local_node);
+  
+  	token = pg_strtok(&length); /* skip :startQual */
+  	local_node->startQual = (List *) nodeRead(true); /* get startQual */
+  
+  	token = pg_strtok(&length); /* skip :parentExpr */
+  	local_node->parentExpr = (List *) nodeRead(true); /* get startQual */
+ 
+  	token = pg_strtok(&length); /* skip :childExpr */
+  	local_node->childExpr = (List *) nodeRead(true); /* get startQual */
+  
+  
+  	token = pg_strtok(&length); /* skip :connop */
+  	token = pg_strtok(&length); /* get connop */
+  	local_node->connOp = atoui(token);
+  
+  	return local_node;
+ }
+  
+ /* ----------------
   *		_readSort
   *
   *	Sort is a subclass of Plan
***************
*** 1013,1018 ****
--- 1079,1114 ----
  }
  
  /* ----------------
+  *		_readFakeVar
+  *
+  *	FakeVar is a subclass of Expr
+  * ----------------
+  */
+ static FakeVar *
+ _readFakeVar(void)
+ {
+ 	FakeVar		   *local_node;
+ 	char	   *token;
+ 	int			length;
+ 
+ 	local_node = makeNode(FakeVar);
+ 
+ 	token = pg_strtok(&length); /* eat :varattno */
+ 	token = pg_strtok(&length); /* get varattno */
+ 	local_node->varattno = atoi(token);
+ 
+ 	token = pg_strtok(&length); /* eat :vartype */
+ 	token = pg_strtok(&length); /* get vartype */
+ 	local_node->vartype = atooid(token);
+ 
+ 	token = pg_strtok(&length); /* eat :vartypmod */
+ 	token = pg_strtok(&length); /* get vartypmod */
+ 	local_node->vartypmod = atoi(token);
+ 
+ 	return local_node;
+ }
+ 
+ /* ----------------
   * _readArrayRef
   *
   * ArrayRef is a subclass of Expr
***************
*** 2272,2277 ****
--- 2368,2379 ----
  		return_value = _readBooleanTest();
  	else if (length == 14 && strncmp(token, "CONSTRAINTTEST", length) == 0)
  		return_value = _readConstraintTest();
+  	else if (length == 4 && strncmp(token, "CONN", length) == 0)
+  		return_value = _readConn();
+  	else if (length == 10 && strncmp(token, "HIERCLAUSE", length) == 0)
+  		return_value = _readHierClause();
+  	else if (length == 7 && strncmp(token, "FAKEVAR", length) == 0)
+  		return_value = _readFakeVar();
  	else
  		elog(ERROR, "badly formatted planstring \"%.10s\"...", token);
  
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/optimizer/path/costsize.c postgresql-7.3rc1-evg/src/backend/optimizer/path/costsize.c
*** postgresql-7.3rc1/src/backend/optimizer/path/costsize.c	Thu Sep  5 01:31:20 2002
--- postgresql-7.3rc1-evg/src/backend/optimizer/path/costsize.c	Thu Nov 28 19:15:21 2002
***************
*** 512,517 ****
--- 512,582 ----
  }
  
  
+ void
+ cost_conn(Path *path, Query *root,
+ 		  List *pathkeys, double tuples, int width)
+ {
+ 	Cost		startup_cost = 0;
+ 	Cost		run_cost = 0;	
+ 	double		nbytes = relation_byte_size(tuples, width);
+ 	long		sortmembytes = SortMem * 1024L;
+ 	double	x,cmpamnt,i,j;
+ 
+ 	/*
+ 	 * We want to be sure the cost of a sort is never estimated as zero,
+  	 * even if passed-in tuple count is zero.  Besides, mustn't do
+  	 * log(0)...
+  	 */
+  	if (tuples < 2.0)
+  		tuples = 2.0;
+  
+  	x = log(tuples)/log(log(tuples));
+  	x = floor(x); //number of runs, slightly overestimated
+ 	
+  	cmpamnt=0.0;
+  	for(i=1;i<x;i++)
+  		for(j=0;j<x-i-1;j++)
+  			cmpamnt+=pow(x,x-j);
+  	cmpamnt=floor(cmpamnt/2);// rought amount of compares = half of worst case
+  	if (cmpamnt < 1.0)
+  		cmpamnt=1.0;
+  	/*
+  	 * CPU costs
+  	 *
+ 	 * assume: cost of comparation + cost of evaluating parent and child expr ==
+ 	 *           3 * cost of comparation 
+ 	 * cost is:	3 * cost of comparation *
+ 	 *            ( amount of compares + check every tuple for being top node )
+ 	 * if HAVING clause present then add cost of qualification of every tuple.
+ 	 * assume we're connected all of input tuples, in real this may be not a true, so
+ 	 * we're estimating the worst case.
+  	 */
+  	startup_cost += 3 * cpu_operator_cost * ( cmpamnt + tuples );
+  	if(root->havingQual){
+  		startup_cost += cpu_operator_cost * tuples;
+  	}
+  
+  	/* disk costs */
+  	if (nbytes > sortmembytes)
+  	{
+  		double		npages = ceil(nbytes / BLCKSZ);
+  		double		npageaccesses;
+  
+  		npageaccesses = 2.0 * npages * x;
+  		/* Assume half are sequential (cost 1), half are not */
+  		startup_cost += npageaccesses *
+  			(1.0 + cost_nonsequential_access(npages)) * 0.5;
+  	}
+  
+  	/*
+  	 * Note: should we bother to assign a nonzero run_cost to reflect the
+  	 * overhead of extracting tuples from the sort result?	Probably not
+  	 * worth worrying about.
+  	 */
+  	path->startup_cost = startup_cost;
+  	path->total_cost = startup_cost + run_cost;
+ }
+  
  /*
   * cost_nestloop
   *	  Determines and returns the cost of joining two relations using the
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/optimizer/plan/createplan.c postgresql-7.3rc1-evg/src/backend/optimizer/plan/createplan.c
*** postgresql-7.3rc1/src/backend/optimizer/plan/createplan.c	Fri Nov 15 06:37:08 2002
--- postgresql-7.3rc1-evg/src/backend/optimizer/plan/createplan.c	Thu Nov 28 20:55:33 2002
***************
*** 833,838 ****
--- 833,839 ----
  		case T_FunctionScan:
  		case T_Material:
  		case T_Sort:
+  		case T_Conn:
  			/* OK, these inner plans support mark/restore */
  			break;
  
***************
*** 1547,1552 ****
--- 1548,1597 ----
  	node->keycount = keycount;
  
  	return node;
+ }
+ 
+ /* called in same terms as make_sort.
+  * copy fileds from HierClause to newly created Conn node,
+  * and calc costs 
+  */
+ Conn *
+ make_conn(Query *root,
+ 			 List *tlist,
+ 			 Plan *lefttree)
+ {
+ 	Conn    *node = makeNode(Conn);
+  	Plan	   *plan = &node->plan;
+  	HierClause *hier = (HierClause *)root->hierClause;
+  
+  	Path		conn_path;		/* dummy for result of cost_conn */
+  
+  	copy_plan_costsize(plan, lefttree); /* only care about copying size */
+  	cost_conn(&conn_path, root, NIL,
+  			  lefttree->plan_rows, lefttree->plan_width);
+  	plan->startup_cost = conn_path.startup_cost + lefttree->startup_cost;
+  	plan->total_cost = conn_path.total_cost + lefttree->total_cost;
+  
+  	plan->plan_rows = lefttree->plan_rows;
+  	plan->plan_width = lefttree->plan_width;
+  	plan->state = (EState *) NULL;
+  	plan->targetlist = tlist;
+  	plan->qual = root->havingQual;
+  	plan->lefttree = lefttree;
+  	plan->righttree = NULL;
+  	node->startQual = hier->startQual;
+ 	
+ 	if(length((List *)hier->parentExpr)>1)
+ 		elog(ERROR,"make_conn: parent expression can't be a list");
+ 	else
+ 		node->parentExpr = lfirst((List *)hier->parentExpr);
+ 		
+ 	if(length((List *)hier->childExpr)>1)
+ 		elog(ERROR,"make_conn: parent expression can't be a list");
+ 	else
+ 		node->childExpr = lfirst((List *)hier->childExpr);
+ 	
+  	node->connOp = hier->connOp;
+  	return node;
  }
  
  /*
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/optimizer/plan/initsplan.c postgresql-7.3rc1-evg/src/backend/optimizer/plan/initsplan.c
*** postgresql-7.3rc1/src/backend/optimizer/plan/initsplan.c	Thu Sep  5 01:31:21 2002
--- postgresql-7.3rc1-evg/src/backend/optimizer/plan/initsplan.c	Thu Nov 28 19:15:21 2002
***************
*** 168,174 ****
  	foreach(temp, vars)
  	{
  		Var		   *var = (Var *) lfirst(temp);
! 		RelOptInfo *rel = find_base_rel(root, var->varno);
  
  		add_var_to_tlist(rel, var);
  
--- 168,179 ----
  	foreach(temp, vars)
  	{
  		Var		   *var = (Var *) lfirst(temp);
! 		RelOptInfo *rel;
! 		/* skip FakeVars */
! 		if(IsA((Node *)var, FakeVar))
! 			continue;
! 		
! 		rel = find_base_rel(root, var->varno);
  
  		add_var_to_tlist(rel, var);
  
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/optimizer/plan/planner.c postgresql-7.3rc1-evg/src/backend/optimizer/plan/planner.c
*** postgresql-7.3rc1/src/backend/optimizer/plan/planner.c	Tue Sep 24 23:38:23 2002
--- postgresql-7.3rc1-evg/src/backend/optimizer/plan/planner.c	Thu Nov 28 19:15:21 2002
***************
*** 39,45 ****
  #define EXPRKIND_TARGET 0
  #define EXPRKIND_WHERE	1
  #define EXPRKIND_HAVING 2
! 
  
  static Node *pull_up_subqueries(Query *parse, Node *jtnode,
  				   bool below_outer_join);
--- 39,45 ----
  #define EXPRKIND_TARGET 0
  #define EXPRKIND_WHERE	1
  #define EXPRKIND_HAVING 2
! #define EXPRKIND_STARTWITH 3
  
  static Node *pull_up_subqueries(Query *parse, Node *jtnode,
  				   bool below_outer_join);
***************
*** 170,175 ****
--- 170,184 ----
  	parse->havingQual = preprocess_expression(parse, parse->havingQual,
  											  EXPRKIND_HAVING);
  
+  	if(parse->hierClause){
+  		((HierClause *)parse->hierClause)->startQual = preprocess_expression(parse, ((HierClause *)parse->hierClause)->startQual,
+  											  EXPRKIND_STARTWITH);
+ 
+  		((HierClause *)parse->hierClause)->parentExpr = preprocess_expression(parse, ((HierClause *)parse->hierClause)->parentExpr,
+  											  EXPRKIND_STARTWITH);
+  		((HierClause *)parse->hierClause)->childExpr = preprocess_expression(parse, ((HierClause *)parse->hierClause)->childExpr,
+  											  EXPRKIND_STARTWITH);		
+  	}
  	/* Also need to preprocess expressions for function RTEs */
  	foreach(lst, parse->rtable)
  	{
***************
*** 210,227 ****
  	 * declared as Node *.	Also note that contain_agg_clause does not
  	 * recurse into sub-selects, which is exactly what we need here.
  	 */
! 	newHaving = NIL;
! 	foreach(lst, (List *) parse->havingQual)
! 	{
! 		Node	   *havingclause = (Node *) lfirst(lst);
  
! 		if (contain_agg_clause(havingclause))
! 			newHaving = lappend(newHaving, havingclause);
! 		else
! 			parse->jointree->quals = (Node *)
! 				lappend((List *) parse->jointree->quals, havingclause);
! 	}
! 	parse->havingQual = (Node *) newHaving;
  
  	/*
  	 * Do the main planning.  If we have an inherited target relation,
--- 219,254 ----
  	 * declared as Node *.	Also note that contain_agg_clause does not
  	 * recurse into sub-selects, which is exactly what we need here.
  	 */
! 	if(!parse->hierClause){
! 		newHaving = NIL;
! 		foreach(lst, (List *) parse->havingQual)
! 		{
! 			Node	   *havingclause = (Node *) lfirst(lst);
  
! 			if (contain_agg_clause(havingclause))
! 				newHaving = lappend(newHaving, havingclause);
! 			else
! 				parse->jointree->quals = (Node *)
! 					lappend((List *) parse->jointree->quals, havingclause);
! 		}
! 		parse->havingQual = (Node *) newHaving;
!  	}else{ /* same as above, but another check */
!  		newHaving = NIL;
!  		foreach(lst, (List *) parse->havingQual)
!  		{
!  			bool res;
!  			Node	   *havingclause = (Node *) lfirst(lst);
!  			
!  			res = contain_fake_column(havingclause);
!  			if (res)
!  				newHaving = lappend(newHaving, havingclause);
!  			else
!  				parse->jointree->quals = (Node *)
!  					lappend((List *) parse->jointree->quals, havingclause);
!  		}
!  		parse->havingQual = (Node *) newHaving;	
!  
!  	}
  
  	/*
  	 * Do the main planning.  If we have an inherited target relation,
***************
*** 371,376 ****
--- 398,416 ----
  				ResolveNew(parse->havingQual,
  						   varno, 0, subtlist, CMD_SELECT, 0);
  
+  			if(parse->hierClause){
+  				((HierClause *)parse->hierClause)->startQual = 
+  						ResolveNew(((HierClause *)parse->hierClause)->startQual,
+  								   varno, 0, subtlist, CMD_SELECT, 0);
+ 
+  				((HierClause *)parse->hierClause)->parentExpr = 
+  						ResolveNew(((HierClause *)parse->hierClause)->parentExpr,
+  								   varno, 0, subtlist, CMD_SELECT, 0);
+  				((HierClause *)parse->hierClause)->childExpr = 
+  						ResolveNew(((HierClause *)parse->hierClause)->childExpr,
+  								   varno, 0, subtlist, CMD_SELECT, 0);
+  			}
+ 			
  			foreach(rt, parse->rtable)
  			{
  				RangeTblEntry *rte = (RangeTblEntry *) lfirst(rt);
***************
*** 495,506 ****
  
  	/*
  	 * Can't pull up a subquery involving grouping, aggregation, sorting,
! 	 * or limiting.
  	 */
  	if (subquery->hasAggs ||
  		subquery->groupClause ||
  		subquery->havingQual ||
  		subquery->sortClause ||
  		subquery->distinctClause ||
  		subquery->limitOffset ||
  		subquery->limitCount)
--- 535,547 ----
  
  	/*
  	 * Can't pull up a subquery involving grouping, aggregation, sorting,
! 	 * limiting or connecting.
  	 */
  	if (subquery->hasAggs ||
  		subquery->groupClause ||
  		subquery->havingQual ||
  		subquery->sortClause ||
+  		subquery->hierClause ||
  		subquery->distinctClause ||
  		subquery->limitOffset ||
  		subquery->limitCount)
***************
*** 1272,1277 ****
--- 1313,1321 ----
  	 * If aggregate is present, insert the Agg node
  	 *
  	 * HAVING clause, if any, becomes qual of the Agg node
+  	 *
+  	 * if there is no hasAggs, then it becomes qual for the
+  	 * Conn node, if it is present
  	 */
  	if (parse->hasAggs)
  	{
***************
*** 1280,1290 ****
  										result_plan);
  		/* Note: Agg does not affect any existing sort order of the tuples */
  	}
- 	else
- 	{
- 		/* If there are no Aggs, we shouldn't have any HAVING qual anymore */
- 		Assert(parse->havingQual == NULL);
- 	}
  
  	/*
  	 * If we were not able to make the plan come out in the right order,
--- 1324,1329 ----
***************
*** 1306,1311 ****
--- 1345,1365 ----
  										   parse->distinctClause);
  	}
  
+  	/*
+  	 * If there is a CONNECT BY clause, add the CONN node
+  	 */
+  	if (parse->hierClause)
+  	{
+  		result_plan = (Plan *) make_connplan(parse, tlist, result_plan);
+  	}
+  	else
+  	{
+  		/* If there are no Aggs and no hierClause , we shouldn't have 
+  		 * any HAVING qual anymore 
+  		 */
+  		if(!parse->hasAggs) Assert(parse->havingQual == NULL);
+  	}
+  
  	/*
  	 * Finally, if there is a LIMIT/OFFSET clause, add the LIMIT node.
  	 */
***************
*** 1370,1376 ****
  	 * If we're not grouping or aggregating, nothing to do here;
  	 * query_planner should receive the unmodified target list.
  	 */
! 	if (!parse->hasAggs && !parse->groupClause && !parse->havingQual)
  		return tlist;
  
  	/*
--- 1424,1430 ----
  	 * If we're not grouping or aggregating, nothing to do here;
  	 * query_planner should receive the unmodified target list.
  	 */
!  	if (!parse->hasAggs && !parse->groupClause && !(parse->havingQual && !parse->hierClause))
  		return tlist;
  
  	/*
***************
*** 1530,1535 ****
--- 1584,1599 ----
  	Assert(keyno > 0);
  
  	return (Plan *) make_sort(parse, sort_tlist, plannode, keyno);
+ }
+ 
+ /*
+  * make_connplan
+  *	  Add a Conn node to implement an explicit CONNECT BY clause.
+  */
+ Plan *
+ make_connplan(Query *parse, List *tlist, Plan *plannode)
+ {
+  	return (Plan *) make_conn(parse, tlist, plannode);
  }
  
  /*
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/optimizer/plan/setrefs.c postgresql-7.3rc1-evg/src/backend/optimizer/plan/setrefs.c
*** postgresql-7.3rc1/src/backend/optimizer/plan/setrefs.c	Thu Sep  5 01:31:21 2002
--- postgresql-7.3rc1-evg/src/backend/optimizer/plan/setrefs.c	Thu Nov 28 19:15:21 2002
***************
*** 23,28 ****
--- 23,30 ----
  #include "optimizer/tlist.h"
  #include "optimizer/var.h"
  #include "parser/parsetree.h"
+ #include "catalog/pg_type.h"
+  
  
  
  typedef struct
***************
*** 162,167 ****
--- 164,178 ----
  			fix_expr_references(plan,
  							  (Node *) ((HashJoin *) plan)->hashclauses);
  			break;
+  		case T_Conn:
+  			set_uppernode_references(plan, (Index) 0);			
+  			fix_expr_references(plan, (Node *) plan->targetlist);
+  			fix_expr_references(plan, (Node *) plan->qual);
+  			fix_expr_references(plan, (Node *)((Conn *) plan)->startQual);
+ 
+  			fix_expr_references(plan, (Node *)((Conn *) plan)->parentExpr);
+  			fix_expr_references(plan, (Node *)((Conn *) plan)->childExpr);
+  			break;
  		case T_Material:
  		case T_Sort:
  		case T_Unique:
***************
*** 361,366 ****
--- 372,395 ----
  									   subvarno,
  									   subplan_targetlist,
  									   tlist_has_non_vars);
+  	if(IsA( plan, Conn)){
+  		((Conn *)plan)->startQual = (List *)
+  			replace_vars_with_subplan_refs((Node *) ((Conn *)plan)->startQual,
+  									   subvarno,
+  									   subplan_targetlist,
+  									   tlist_has_non_vars);
+ 
+  		((Conn *)plan)->parentExpr = (List *)
+  			replace_vars_with_subplan_refs((Node *) ((Conn *)plan)->parentExpr,
+  									   subvarno,
+  									   subplan_targetlist,
+  									   tlist_has_non_vars);
+  		((Conn *)plan)->childExpr = (List *)
+  			replace_vars_with_subplan_refs((Node *) ((Conn *)plan)->childExpr,
+  									   subvarno,
+  									   subplan_targetlist,
+  									   tlist_has_non_vars);
+  	}
  }
  
  /*
***************
*** 508,514 ****
  {
  	if (node == NULL)
  		return NULL;
! 	if (IsA(node, Var))
  	{
  		Var		   *var = (Var *) node;
  		Resdom	   *resdom;
--- 537,552 ----
  {
  	if (node == NULL)
  		return NULL;
! 	if (IsA(node, FakeVar))
! 	{
! 		/* FakeVar isn't present on subplan so just copy it*/
! 		FakeVar		   *var = (FakeVar *) node;
! 		Var		   *newvar;
! 
! 		newvar = (FakeVar *) copyObject(var);
! 		return (Node *) newvar;
! 	}
! 	else if (IsA(node, Var))
  	{
  		Var		   *var = (Var *) node;
  		Resdom	   *resdom;
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/optimizer/plan/subselect.c postgresql-7.3rc1-evg/src/backend/optimizer/plan/subselect.c
*** postgresql-7.3rc1/src/backend/optimizer/plan/subselect.c	Thu Sep  5 01:31:21 2002
--- postgresql-7.3rc1-evg/src/backend/optimizer/plan/subselect.c	Thu Nov 28 19:15:21 2002
***************
*** 355,360 ****
--- 355,361 ----
  				case T_Material:
  				case T_FunctionScan:
  				case T_Sort:
+ 				case T_Conn:
  
  					/*
  					 * Don't add another Material node if there's one
***************
*** 699,704 ****
--- 700,706 ----
  		case T_SetOp:
  		case T_Limit:
  		case T_Group:
+ 		case T_Conn:
  			break;
  
  		default:
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/optimizer/util/clauses.c postgresql-7.3rc1-evg/src/backend/optimizer/util/clauses.c
*** postgresql-7.3rc1/src/backend/optimizer/util/clauses.c	Wed Sep 11 19:48:54 2002
--- postgresql-7.3rc1-evg/src/backend/optimizer/util/clauses.c	Thu Nov 28 19:15:21 2002
***************
*** 56,61 ****
--- 56,62 ----
  static bool contain_volatile_functions_walker(Node *node, void *context);
  static Node *eval_const_expressions_mutator(Node *node, void *context);
  static Expr *simplify_op_or_func(Expr *expr, List *args);
+ static bool contain_fake_column_walker(Node *node, void *context);
  
  
  Expr *
***************
*** 453,458 ****
--- 454,488 ----
  								  (void *) listptr);
  }
  
+ /*****************************************************************************
+  *		Connect by manipulation
+  *****************************************************************************/
+ 
+ /*
+  * contain_fake_var
+  *	  Recursively search for fake var nodes within a clause.
+  *
+  *	  Returns true if any such var found.
+  */
+ bool
+ contain_fake_column(Node *clause)
+ {
+ 	bool result;
+  
+  	result = contain_fake_column_walker(clause, NULL);
+  	return result;
+ }
+  
+ static bool
+ contain_fake_column_walker(Node *node, void *context)
+ {
+  	if (node == NULL)
+  		return false;
+  	if (IsA(node, FakeVar))
+  		return true;			/* abort the tree traversal and return
+  								 * true */
+  	return expression_tree_walker(node, contain_fake_column_walker, context);
+ }
  
  /*****************************************************************************
   *		Support for expressions returning sets
***************
*** 1007,1012 ****
--- 1037,1044 ----
  		Var		   *var = (Var *) lfirst(i);
  		List	   *vi;
  
+ 		if(IsA(lfirst(i), FakeVar)) /* skip FakeVars, because they're not fetched from tables */
+ 			continue;
  		if (!intMember(var->varno, varno_list))
  			varno_list = lconsi(var->varno, varno_list);
  		foreach(vi, var_list)
***************
*** 1824,1829 ****
--- 1856,1862 ----
  	{
  		case T_Const:
  		case T_Var:
+ 		case T_FakeVar:
  		case T_Param:
  		case T_RangeTblRef:
  			/* primitive node types with no subnodes */
***************
*** 1977,1982 ****
--- 2010,2026 ----
  					return true;
  			}
  			break;
+  		case T_HierClause:
+  			{
+  				HierClause *hier = (HierClause *)node;
+  				if (walker(hier->startQual, context))
+  					return true;
+  				if (walker(hier->parentExpr, context))
+  					return true;
+  				if (walker(hier->childExpr, context))
+  					return true;
+  			}
+  			break;
  		default:
  			elog(ERROR, "expression_tree_walker: Unexpected node type %d",
  				 nodeTag(node));
***************
*** 2018,2023 ****
--- 2062,2070 ----
  		return true;
  	if (walker(query->havingQual, context))
  		return true;
+  	if (walker(query->hierClause, context))
+  		return true;
+  	
  	foreach(rt, query->rtable)
  	{
  		RangeTblEntry *rte = (RangeTblEntry *) lfirst(rt);
***************
*** 2134,2140 ****
  	switch (nodeTag(node))
  	{
  		case T_Const:
! 		case T_Var:
  		case T_Param:
  		case T_RangeTblRef:
  			/* primitive node types with no subnodes */
--- 2181,2188 ----
  	switch (nodeTag(node))
  	{
  		case T_Const:
! 		case T_FakeVar:
! 		case T_Var:		
  		case T_Param:
  		case T_RangeTblRef:
  			/* primitive node types with no subnodes */
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/optimizer/util/var.c postgresql-7.3rc1-evg/src/backend/optimizer/util/var.c
*** postgresql-7.3rc1/src/backend/optimizer/util/var.c	Wed Sep 11 19:48:54 2002
--- postgresql-7.3rc1-evg/src/backend/optimizer/util/var.c	Thu Nov 28 19:15:21 2002
***************
*** 255,260 ****
--- 255,262 ----
  {
  	if (node == NULL)
  		return false;
+ 	if(IsA(node, FakeVar)) /* fake vars are vars too */
+ 		return true;
  	if (IsA(node, Var))
  	{
  		if (((Var *) node)->varlevelsup == 0)
***************
*** 301,306 ****
--- 303,316 ----
  	{
  		if (((Var *) node)->varlevelsup == 0 || context->includeUpperVars)
  			context->varlist = lappend(context->varlist, node);
+ 		return false;
+ 	}
+ 	if (IsA(node, FakeVar))
+ 	{
+ 		/* same as Var but because of FakeVar->varlevelsup == 0 is always true (ideologically),
+ 		 * always append
+ 		 */
+ 		context->varlist = lappend(context->varlist, node);
  		return false;
  	}
  	return expression_tree_walker(node, pull_var_clause_walker,
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/parser/analyze.c postgresql-7.3rc1-evg/src/backend/parser/analyze.c
*** postgresql-7.3rc1/src/backend/parser/analyze.c	Tue Oct 22 03:06:19 2002
--- postgresql-7.3rc1-evg/src/backend/parser/analyze.c	Thu Nov 28 19:15:21 2002
***************
*** 1659,1664 ****
--- 1659,1667 ----
  	/* transform WHERE */
  	qual = transformWhereClause(pstate, stmt->whereClause);
  
+   /* transform CONNECT BY , START WITH clause */
+ 	qry->hierClause=transformHierClause(pstate,qry,stmt->hierClause);
+ 
  	/*
  	 * Initial processing of HAVING clause is just like WHERE clause.
  	 * Additional work will be done in optimizer/plan/planner.c.
***************
*** 1683,1689 ****
  
  	qry->hasSubLinks = pstate->p_hasSubLinks;
  	qry->hasAggs = pstate->p_hasAggs;
! 	if (pstate->p_hasAggs || qry->groupClause || qry->havingQual)
  		parseCheckAggregates(pstate, qry, qual);
  
  	qry->rtable = pstate->p_rtable;
--- 1686,1692 ----
  
  	qry->hasSubLinks = pstate->p_hasSubLinks;
  	qry->hasAggs = pstate->p_hasAggs;
!  	if (pstate->p_hasAggs || qry->groupClause || (qry->havingQual && !qry->hierClause) )
  		parseCheckAggregates(pstate, qry, qual);
  
  	qry->rtable = pstate->p_rtable;
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/parser/gram.y postgresql-7.3rc1-evg/src/backend/parser/gram.y
*** postgresql-7.3rc1/src/backend/parser/gram.y	Sat Nov  2 22:41:21 2002
--- postgresql-7.3rc1-evg/src/backend/parser/gram.y	Thu Nov 28 19:15:21 2002
***************
*** 311,316 ****
--- 311,318 ----
  %type <list>	constraints_set_list
  %type <boolean> constraints_set_mode
  
+ %type <node>  startwith_clause
+ %type <list>  connectby_clause hier_clause 
  
  /*
   * If you make any token changes, update the keyword table in
***************
*** 329,336 ****
  	CACHE CALLED CASCADE CASE CAST CHAIN CHAR_P
  	CHARACTER CHARACTERISTICS CHECK CHECKPOINT CLASS CLOSE
  	CLUSTER COALESCE COLLATE COLUMN COMMENT COMMIT
! 	COMMITTED CONSTRAINT CONSTRAINTS CONVERSION_P CONVERT COPY CREATE CREATEDB
! 	CREATEUSER CROSS CURRENT_DATE CURRENT_TIME
  	CURRENT_TIMESTAMP CURRENT_USER CURSOR CYCLE
  
  	DATABASE DAY_P DEALLOCATE DEC DECIMAL DECLARE DEFAULT
--- 331,338 ----
  	CACHE CALLED CASCADE CASE CAST CHAIN CHAR_P
  	CHARACTER CHARACTERISTICS CHECK CHECKPOINT CLASS CLOSE
  	CLUSTER COALESCE COLLATE COLUMN COMMENT COMMIT
! 	COMMITTED CONNECT CONSTRAINT CONSTRAINTS CONVERSION_P CONVERT 
! 	COPY CREATE CREATEDB CREATEUSER CROSS CURRENT_DATE CURRENT_TIME
  	CURRENT_TIMESTAMP CURRENT_USER CURSOR CYCLE
  
  	DATABASE DAY_P DEALLOCATE DEC DECIMAL DECLARE DEFAULT
***************
*** 4232,4238 ****
  simple_select:
  			SELECT opt_distinct target_list
  			into_clause from_clause where_clause
! 			group_clause having_clause
  				{
  					SelectStmt *n = makeNode(SelectStmt);
  					n->distinctClause = $2;
--- 4234,4240 ----
  simple_select:
  			SELECT opt_distinct target_list
  			into_clause from_clause where_clause
! 			group_clause hier_clause having_clause
  				{
  					SelectStmt *n = makeNode(SelectStmt);
  					n->distinctClause = $2;
***************
*** 4242,4248 ****
  					n->fromClause = $5;
  					n->whereClause = $6;
  					n->groupClause = $7;
! 					n->havingClause = $8;
  					$$ = (Node *)n;
  				}
  			| select_clause UNION opt_all select_clause
--- 4244,4251 ----
  					n->fromClause = $5;
  					n->whereClause = $6;
  					n->groupClause = $7;
!  					n->havingClause = $9;
!  					n->hierClause = $8;
  					$$ = (Node *)n;
  				}
  			| select_clause UNION opt_all select_clause
***************
*** 4264,4269 ****
--- 4267,4295 ----
  			| /*EMPTY*/								{ $$ = NULL; }
  		;
  
+ hier_clause: connectby_clause startwith_clause 
+  			{ 
+  				$$ = makeList2($1,$2); 
+  			}
+  		| startwith_clause connectby_clause 
+  			{ 
+  				$$ = makeList2($2,$1); 
+  			}
+  		| /* EMPTY */ { $$ = NULL; }
+  		;
+  
+ connectby_clause: CONNECT BY PRIOR c_expr all_Op c_expr
+ 			{ 
+  			  $$ = makeList3($4,makeString($5),$6); 
+  			}
+ 		| CONNECT BY a_expr qual_Op PRIOR a_expr
+ 			{ 
+  			  $$ = makeList3($6,makeString($4),$3); 
+  			}
+  		;
+ startwith_clause: START WITH a_expr { $$ = $3; }
+  		;
+  		
  /*
   * Redundancy here is needed to avoid shift/reduce conflicts,
   * since TEMP is not a reserved word.  See also OptTemp.
***************
*** 7074,7080 ****
  			| PENDANT
  			| PRECISION
  			| PREPARE
- 			| PRIOR
  			| PRIVILEGES
  			| PROCEDURAL
  			| PROCEDURE
--- 7100,7105 ----
***************
*** 7102,7108 ****
  			| SHOW
  			| SIMPLE
  			| STABLE
- 			| START
  			| STATEMENT
  			| STATISTICS
  			| STDIN
--- 7127,7132 ----
***************
*** 7239,7244 ****
--- 7263,7269 ----
  			| CHECK
  			| COLLATE
  			| COLUMN
+ 			| CONNECT
  			| CONSTRAINT
  			| CREATE
  			| CURRENT_DATE
***************
*** 7279,7288 ****
--- 7304,7315 ----
  			| ORDER
  			| PLACING
  			| PRIMARY
+ 			| PRIOR
  			| REFERENCES
  			| SELECT
  			| SESSION_USER
  			| SOME
+ 			| START
  			| TABLE
  			| THEN
  			| TO
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/parser/keywords.c postgresql-7.3rc1-evg/src/backend/parser/keywords.c
*** postgresql-7.3rc1/src/backend/parser/keywords.c	Thu Sep 19 02:35:22 2002
--- postgresql-7.3rc1-evg/src/backend/parser/keywords.c	Thu Nov 28 19:15:21 2002
***************
*** 79,84 ****
--- 79,85 ----
  	{"comment", COMMENT},
  	{"commit", COMMIT},
  	{"committed", COMMITTED},
+  	{"connect",CONNECT},
  	{"constraint", CONSTRAINT},
  	{"constraints", CONSTRAINTS},
  	{"conversion", CONVERSION_P},
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/parser/parse_clause.c postgresql-7.3rc1-evg/src/backend/parser/parse_clause.c
*** postgresql-7.3rc1/src/backend/parser/parse_clause.c	Thu Sep 19 02:35:22 2002
--- postgresql-7.3rc1-evg/src/backend/parser/parse_clause.c	Thu Nov 28 19:15:21 2002
***************
*** 37,44 ****
  #define ORDER_CLAUSE 0
  #define GROUP_CLAUSE 1
  #define DISTINCT_ON_CLAUSE 2
! 
! static char *clauseText[] = {"ORDER BY", "GROUP BY", "DISTINCT ON"};
  
  static void extractRemainingColumns(List *common_colnames,
  						List *src_colnames, List *src_colvars,
--- 37,45 ----
  #define ORDER_CLAUSE 0
  #define GROUP_CLAUSE 1
  #define DISTINCT_ON_CLAUSE 2
! #define CONNECT_CLAUSE 3
!   
! static char *clauseText[] = {"ORDER BY", "GROUP BY", "DISTINCT ON","CONNECT BY"};
  
  static void extractRemainingColumns(List *common_colnames,
  						List *src_colnames, List *src_colvars,
***************
*** 1158,1163 ****
--- 1159,1234 ----
  	}
  
  	return glist;
+ }
+ 
+ /*
+  * transformConnectClause -
+  *	  transform a Connect By clause
+  *
+  */
+ Node *
+ transformHierClause(ParseState *pstate, Query *qry, List *hier)
+ {
+ 	List	   *clist = NIL;	
+ 	TargetEntry *tle;
+ 	HierClause *node;
+ 	List *tlist=qry->targetList;
+ 	FakeVar *var;
+ 	
+ 	if(!hier) return NULL;
+  	
+  	node=makeNode(HierClause);
+  	
+  	clist=lfirst(hier);//get conn_list
+ 	
+  	/* add a field '_level_', in result there will be deepness of tuple 
+ 	 * in result tree.
+ 	 * must be added before tranformation parent/child Expr,
+ 	 * because in they transformation process may be added some resjunk
+ 	 * var nodes to tlist, scanRTEForColumn() not supports mix of junk/non-junk
+ 	 * nodes.
+  	 */
+ 	var = makeNode(FakeVar);
+ 	var->varattno = 0;
+ 	var->vartype = INT4OID;
+ 	var->vartypmod = -1;	
+  	
+  	tle = transformTargetEntry(pstate, (Node *)var, NULL, "_level_", false);
+  	lappend(tlist, tle);
+ 
+ 	/* transform parent and child expressions */
+ 	pstate->p_curTlist = tlist; /* also see in parser/parse_node.h */
+ 	pstate->p_addTle = true;    /* we may add to tagetlist some junk attributes,
+ 															 * which needed by this expressions but not present in 
+ 															 * user given targetlist. only vars is added */
+ 	node->parentExpr = transformExpr(pstate,lfirst(clist));
+ 	node->childExpr = transformExpr(pstate,lfirst(lnext(lnext(clist))));//lthird
+ 	pstate->p_addTle = false;
+ 	
+ 	/* fetch operatior OID for connecting tuples */
+  	node->connOp = compatible_oper_opid(makeList1(lsecond(clist)),
+ 					exprType(node->parentExpr),
+  					exprType(node->childExpr),
+ 					true);
+ 
+  	if(node->connOp == InvalidOid){
+  	  elog(ERROR,"CONNECT BY: can't find operator '%s' for connecting on chosen fields",(char *)lsecond(clist));
+  	}
+ 	
+ 	/* transform StartQual, see note above, on parent/child Expr*/
+ 	pstate->p_addTle = true;
+ 	node->startQual=transformWhereClause(pstate,lsecond(hier));
+ 	pstate->p_addTle = false;
+ 	pstate->p_curTlist = NULL;// clear
+  	
+  	pstate->p_hierLevelColIdx = tle->resdom->resno;
+ 	
+ 	/* transformExpr return to us our's pointer, so we can to set
+ 	 * attno here, for FakeVar attno is the number in targetlist
+ 	 */
+ 	var->varattno = tle->resdom->resno;
+  
+  	return (Node *)node;
  }
  
  /*
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/parser/parse_expr.c postgresql-7.3rc1-evg/src/backend/parser/parse_expr.c
*** postgresql-7.3rc1/src/backend/parser/parse_expr.c	Thu Sep 19 02:35:22 2002
--- postgresql-7.3rc1-evg/src/backend/parser/parse_expr.c	Thu Nov 28 20:47:26 2002
***************
*** 633,640 ****
  			 * taking a conservative approach, and only accepting node
  			 * types that are demonstrably necessary to accept.
  			 *********************************************/
- 		case T_Expr:
  		case T_Var:
  		case T_Const:
  		case T_Param:
  		case T_Aggref:
--- 633,641 ----
  			 * taking a conservative approach, and only accepting node
  			 * types that are demonstrably necessary to accept.
  			 *********************************************/
  		case T_Var:
+ 		case T_Expr:
+ 		case T_FakeVar:
  		case T_Const:
  		case T_Param:
  		case T_Aggref:
***************
*** 662,669 ****
--- 663,696 ----
  static Node *
  transformIndirection(ParseState *pstate, Node *basenode, List *indirection)
  {
+ 	TargetEntry *tle;
+ 	List *tl;
+ 	bool found;
+ 	
  	if (indirection == NIL)
+ 	{
+ 		/* If node is Var, and tle addition flag is set then we must add var to targetlist
+ 		 * if it's not presen in tlist
+ 		 * ( also see parser/parse_node.h, parser/parse_clause.c:transformHierClause())
+ 		 */
+ 		if(IsA(basenode, Var) && pstate->p_addTle){
+ 			found = false;
+ 			/* search for var */
+ 			foreach(tl, pstate->p_curTlist){
+ 				tle = lfirst(tl);
+ 				if(equal(basenode,tle->expr)){
+ 					found = true;
+ 					break;
+ 				}
+ 			}
+ 			/* if not fount, make tle and add to tlist */
+ 			if(!found){
+ 				tle = transformTargetEntry(pstate,basenode,basenode,NULL,true);
+ 				lappend(pstate->p_curTlist,tle);
+ 			}
+ 		}
  		return basenode;
+ 	}
  	return (Node *) transformArraySubscripts(pstate,
  											 basenode,
  											 exprType(basenode),
***************
*** 740,748 ****
--- 767,788 ----
  						rv->inhOpt = INH_DEFAULT;
  						node = (Node *) rv;
  					}
+ 					else if (pstate->p_hierLevelColIdx!=0 && !strcmp(name,"_level_"))
+ 					{
+ 						/* If nothing of above and present hier clause and name is '_level_'
+ 						 * then make fake variable 
+ 						 */
+ 				 		FakeVar *var = makeNode(FakeVar);
+  		
+ 				 		var->varattno = pstate->p_hierLevelColIdx;
+ 				 		var->vartype = INT4OID;
+ 						var->vartypmod = -1;
+ 				 		node = (Node *)var;
+ 				 	}
  					else
  						elog(ERROR, "Attribute \"%s\" not found", name);
  				}
+ 
  				break;
  			}
  		case 2:
***************
*** 875,880 ****
--- 915,923 ----
  
  	switch (nodeTag(expr))
  	{
+ 		case T_FakeVar:
+ 			type = ((FakeVar *) expr)->vartype;
+ 			break;		
  		case T_Var:
  			type = ((Var *) expr)->vartype;
  			break;
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/utils/adt/ruleutils.c postgresql-7.3rc1-evg/src/backend/utils/adt/ruleutils.c
*** postgresql-7.3rc1/src/backend/utils/adt/ruleutils.c	Fri Sep 20 04:40:56 2002
--- postgresql-7.3rc1-evg/src/backend/utils/adt/ruleutils.c	Thu Nov 28 19:15:21 2002
***************
*** 1929,1934 ****
--- 1929,1939 ----
  			get_const_expr((Const *) node, context);
  			break;
  
+ 		case T_FakeVar:
+ 			{
+ 				appendStringInfo(buf, "FakeVar");
+ 			}
+ 			break;
  		case T_Var:
  			{
  				Var		   *var = (Var *) node;
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/utils/sort/Makefile postgresql-7.3rc1-evg/src/backend/utils/sort/Makefile
*** postgresql-7.3rc1/src/backend/utils/sort/Makefile	Thu Aug 31 21:10:59 2000
--- postgresql-7.3rc1-evg/src/backend/utils/sort/Makefile	Thu Nov 28 19:15:21 2002
***************
*** 12,18 ****
  top_builddir = ../../../..
  include $(top_builddir)/src/Makefile.global
  
! OBJS = logtape.o tuplesort.o tuplestore.o
  
  all: SUBSYS.o
  
--- 12,18 ----
  top_builddir = ../../../..
  include $(top_builddir)/src/Makefile.global
  
! OBJS = logtape.o tuplesort.o tuplestore.o tupleconn.o
  
  all: SUBSYS.o
  
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/backend/utils/sort/tupleconn.c postgresql-7.3rc1-evg/src/backend/utils/sort/tupleconn.c
*** postgresql-7.3rc1/src/backend/utils/sort/tupleconn.c	Thu Jan  1 04:00:00 1970
--- postgresql-7.3rc1-evg/src/backend/utils/sort/tupleconn.c	Thu Nov 28 19:15:21 2002
***************
*** 0 ****
--- 1,938 ----
+ /*-------------------------------------------------------------------------
+  *
+  * tupleconn.c
+  *	  Routines for tuple connecting.
+  *
+  * Based on tuplestore.c, tuplesort.c. (c) by Evgen Potemkin evgent@terminal.ru, 11.2002
+  * 
+  *-------------------------------------------------------------------------
+  */
+ 
+ #include "postgres.h"
+ 
+ #include "access/heapam.h"
+ #include "storage/buffile.h"
+ #include "utils/tupleconn.h"
+ #include "utils/tuplesort.h"
+ 
+ #include "access/nbtree.h"
+ #include "catalog/catname.h"
+ #include "catalog/pg_amop.h"
+ #include "catalog/pg_amproc.h"
+ #include "catalog/pg_operator.h"
+ #include "miscadmin.h"
+ #include "utils/fmgroids.h"
+ #include "utils/lsyscache.h"
+ #include "utils/syscache.h"
+ 
+ #include "nodes/execnodes.h"
+ #include "executor/executor.h"
+ /*
+  * Possible states of a Tupleconn object.	These denote the states that
+  * persist between calls of Tupleconn routines.
+  */
+ typedef enum
+ {
+ 	TCS_INITIAL,				/* Loading tuples; still within memory
+ 								 * limit */
+ 	TCS_WRITEFILE,				/* Loading tuples; writing to temp file */
+ 	TCS_READMEM,				/* Reading tuples; entirely in memory */
+ 	TCS_READFILE				/* Reading tuples from temp file */
+ } TupConnStatus;
+ 
+ 
+ /* 
+  * connection tree 
+  */
+ typedef struct _ConnectTree ConnectTree;
+ struct _ConnectTree{
+   ConnectTree *next; /* next sibling*/
+   ConnectTree *prev; /* previous sibling */
+   ConnectTree *sub;  /* child */
+   ConnectTree *sup;  /* parent */
+   int  level; /* deep of this node, starts from 1*/
+   int  tuple; /* index of tuple in memtuples/used/out_store */
+ };
+ 
+ /*
+  * Information about tuples stored in buffile
+  */
+ typedef struct OutStore{
+ 	int fileno;	/* BufFile fileno */
+ 	long offs;	/* BufFile offs */
+ 	int len;		/* tuple length*/
+ 	HeapTuple tup;	/* in-memory tuple copy, NULL if none */
+ }OutStore;
+ /*
+  * Private state of a Tupleconn operation.
+  */
+ struct Tupleconnstate
+ {
+ 	TupConnStatus status;		/* enumerated value as shown above */
+ 	long		availMem;		/* remaining memory available, in bytes */
+ 	BufFile    *myfile;			/* underlying file, or NULL if none */
+ 
+ 	/*
+ 	 * Tree and supporting stuff, explanation at _performconn function 
+ 	 */
+   ConnectTree *head; /* head of the result list */
+ 	ConnectTree **pnt;  /* parent nodes on curent level*/
+ 	int pntlast;	/* actual amount of parents on current level */
+ 	int pntsize;	/* size of pnt array */
+ 	ConnectTree **chld; /* array of child nodes */
+ 	int chldlast; /* actual amount of childs on current level */
+ 	int chldsize;	/* size of chld array */
+ 	ConnectTree *last; /* last added/fetched node */	
+   bool skip_node; /* used in gettuple, mean don't fall to ->sub because 
+ 	                * we're going from that*/
+ 	char *used;	/* array of flags - was tuple (connected already|is null) or not */
+   TupleDesc tupDesc;
+ 	AttrNumber level; /* tuple's '_level_' attribute number */
+ 	FmgrInfo connFn; /* comparation function */
+ 	
+ 	OutStore *out_store; /* array of info about tuples in buffile for faster random access */
+ 
+ 	void	   *(*copytup) (Tupleconnstate *state, void *tup);
+ 	void		(*writetup) (Tupleconnstate *state, void *tup, bool free, int idx);
+ 	void	   *(*readtup) (Tupleconnstate *state, int idx);
+ 
+ 	void	  **memtuples;		/* array of pointers to palloc'd tuples */
+ 	int			memtupsize;		/* allocated length of memtuples array */
+ 
+ 	int			tupcount;	/* number of tuples currently present */
+ 
+ 	bool		eof_reached;	/* reached EOF (needed for cursors) */
+ 
+ 	/* markpos_xxx holds marked position for mark and restore */
+ 	bool		markpos_eof;	/* saved "eof_reached" */
+ 	ConnectTree *markpos_last; /* saved "last" pointer */
+ 	bool		markpos_skip;	/* saved "skip" flag */
+ };
+ 
+ 
+ #define COPYTUP(state,tup)	((*(state)->copytup) (state, tup))
+ #define WRITETUP(state,tup,free,idx) ((*(state)->writetup) (state, tup, free, idx))
+ #define READTUP(state,idx)	((*(state)->readtup) (state, idx))
+ #define LACKMEM(state)		((state)->availMem < 0)
+ #define USEMEM(state,amt)	((state)->availMem -= (amt))
+ #define FREEMEM(state,amt)	((state)->availMem += (amt))
+ 
+ 
+ 
+ static Tupleconnstate *tupleconn_begin_common(int maxKBytes);
+ static void dumptuples(Tupleconnstate *state);
+ 
+ static void *copytup_heap(Tupleconnstate *state, void *tup);
+ static void writetup_heap(Tupleconnstate *state, void *tup, bool free, int idx);
+ static void *readtup_heap(Tupleconnstate *state, int idx);
+ void SelectConnFunction(Oid sortOperator,RegProcedure *sortFunction,
+ 				   SortFunctionKind *kind);
+ void *setlevel(Tupleconnstate * state,int level, void *tuple, bool free_it);
+ 
+ ConnectTree *add_next(ConnectTree *tr,int tup,int level);
+ ConnectTree *add_sub(ConnectTree *tr,int tup,int level);
+ 
+ 
+ /*
+  * Init all
+  */
+ static Tupleconnstate *
+ tupleconn_begin_common(int maxKBytes)
+ {
+ 	Tupleconnstate *state;
+ 
+ 	state = (Tupleconnstate *) palloc(sizeof(Tupleconnstate));
+ 
+ 	MemSet((char *) state, 0, sizeof(Tupleconnstate));
+ 
+ 	state->status = TCS_INITIAL;
+ 	state->availMem = maxKBytes * 1024L;
+ 	state->myfile = NULL;
+ 	state->out_store = NULL;
+ 	
+ 	state->tupcount = 0;
+ 	if (maxKBytes > 0)
+ 		state->memtupsize = 1024;		/* initial guess */
+ 	else
+ 		state->memtupsize = 1;	/* won't really need any space */
+ 	state->memtuples = (void **) palloc(state->memtupsize * sizeof(void *));
+   state->used = (char *) palloc(state->memtupsize * sizeof(char));
+ 	
+   state->pntlast=0;
+   state->chldlast=0;
+ 	state->pntsize=1024;		/* initial guess */
+ 	state->chldsize=1024;		/* initial guess */
+ 	state->pnt=(ConnectTree **)palloc(state->pntsize * sizeof(ConnectTree *));
+ 	state->chld=(ConnectTree **)palloc(state->pntsize * sizeof(ConnectTree *));
+ 	return state;
+ }
+ 
+ Tupleconnstate *
+ tupleconn_begin_heap(int maxKBytes, TupleDesc tupDesc, Oid connOp)
+ {
+   RegProcedure connFunc;
+ 	SortFunctionKind connFnKind;
+ 	
+ 	Tupleconnstate *state = tupleconn_begin_common(maxKBytes);
+ 
+ 	state->copytup = copytup_heap;
+ 	state->writetup = writetup_heap;
+ 	state->readtup = readtup_heap;
+ 
+   state->tupDesc=tupDesc;
+ 	
+ 	SelectConnFunction(connOp, &connFunc,
+ 						   &connFnKind);
+ 	if(connFnKind != SORTFUNC_CMP){
+ 	  elog(ERROR,"tupleconn: Can't find suitable function for comparation");
+ 	}
+ 	fmgr_info(connFunc,&state->connFn);
+ 	return state;
+ }
+ 
+ /*
+  * tupleconn_end
+  *
+  *	Release resources and clean up.
+  */
+ void
+ tupleconn_end(Tupleconnstate *state)
+ {
+ 	int			i;
+ 	ConnectTree *old;
+ 
+ 	/* common frees */
+ 	pfree(state->pnt);
+ 	pfree(state->chld);
+ 	pfree(state->used);
+ 	
+ 	/* free tree, if any */
+ 	state->last = state->head;
+ 	if(state->last){
+ 		state->skip_node = false;
+   	 while(1) {
+    		if(!state->skip_node && state->last->sub) state->last=state->last->sub;
+ 	   	else if(state->last->next){
+ 				old = state->last;
+ 				state->last=state->last->next;
+ 				state->skip_node=false;
+ 				pfree(old);
+ 			}	else if(state->last->sup){
+ 				old = state->last;
+ 		  	state->last=state->last->sup;
+ 				state->skip_node=true;
+ 				pfree(old);
+ 			}	else {
+ 		  	break;
+ 			}
+   	}
+ 		pfree(state->last);
+ 		state->head = state->last = NULL;
+ 	}
+ 
+ 	/* free tuples, from out_store or memtuples*/
+ 	if (state->myfile){
+ 		BufFileClose(state->myfile);
+ 		for(i=0;i<state->tupcount;i++){
+ 			if(state->out_store[i].tup!=NULL){
+ 				pfree(state->out_store[i].tup);
+ 				FREEMEM(state,state->out_store[i].len);
+ 				state->out_store[i].tup=NULL;
+ 			}
+ 		}
+ 	}else{
+ 		for (i = 0; i < state->tupcount; i++)
+ 			pfree(state->memtuples[i]);
+ 		pfree(state->memtuples);
+ 	}
+ }
+ 
+ /*
+  * Accept one tuple while collecting input data.
+  *
+  * Note that the input tuple is always copied; the caller need not save it.
+  */
+ void
+ tupleconn_puttuple(Tupleconnstate *state, void *tuple,int head)
+ {
+ 	/*
+ 	 * Copy the tuple.	(Must do this even in WRITEFILE case.)
+ 	 */
+ 	tuple = COPYTUP(state, tuple);
+ 
+ 	/* common thing */
+   if(head){//it's a head tuple,add it to head list
+     state->last = add_next(state->last,state->tupcount,1);
+     if(!state->head) state->head = state->last;
+     state->pnt[state->pntlast++] = state->last;
+ 		state->used[state->tupcount] = 1;
+   }else{
+ 	  state->used[state->tupcount] = 0;
+ 	}
+ 	switch (state->status)
+ 	{
+ 		case TCS_INITIAL:
+ 
+ 			/*
+ 			 * Stash the tuple in the in-memory array.
+ 			 */
+ 			if (state->tupcount >= state->memtupsize)
+ 			{
+ 				/* Grow the arrays as needed. */
+ 				state->memtupsize *= 2;
+ 				state->memtuples = (void **)
+ 					repalloc(state->memtuples,
+ 							 state->memtupsize * sizeof(void *));
+ 				state->used = (char *)
+ 					repalloc(state->used,
+ 							 state->memtupsize * sizeof(char));
+ 			}
+ 			state->memtuples[state->tupcount++] = tuple;
+ 
+ 			/*
+ 			 * Done if we still fit in available memory.
+ 			 */
+ 			if (!LACKMEM(state))
+ 				return;
+ 
+ 			/*
+ 			 * Nope; time to switch to tape-based operation.
+ 			 */
+ 			state->myfile = BufFileCreateTemp();			
+ 			state->status = TCS_WRITEFILE;
+ 			state->out_store = palloc(state->memtupsize * sizeof(OutStore));
+ 			dumptuples(state);
+ 			break;
+ 		case TCS_WRITEFILE:
+ 			if (state->tupcount >= state->memtupsize)
+ 			{
+ 				/* Grow the arrays as needed. */
+ 				state->memtupsize *= 2;
+ 				state->out_store = (OutStore *)
+ 					repalloc(state->out_store,
+ 							 state->memtupsize * sizeof(void *));
+ 				state->used = (char *)
+ 					repalloc(state->used,
+ 							 state->memtupsize * sizeof(char));
+ 			}
+ 			WRITETUP(state, tuple,!head,state->tupcount++);
+ 			
+ 			break;
+ 		default:
+ 			elog(ERROR, "tupleconn_puttuple: invalid state (internal error)");
+ 			break;
+ 	}
+ }
+ 
+ /*
+  * All tuples have been provided; finish writing.
+  */
+ void
+ tupleconn_donestoring(Tupleconnstate *state)
+ {
+ 	switch (state->status)
+ 	{
+ 		case TCS_INITIAL:
+ 			/*
+ 			 * We were able to accumulate all the tuples within the
+ 			 * allowed amount of memory.  Just set up to connect and scan them.
+ 			 */
+ 			state->status = TCS_READMEM;
+ 			break;
+ 		case TCS_WRITEFILE:
+ 			/*
+ 			 * Set up for connecting/reading from tape.
+ 			 */
+ 			state->status = TCS_READFILE;
+ 			break;
+ 		default:
+ 			elog(ERROR, "tupleconn_donestoring: invalid state ");
+ 			break;
+ 	}
+ 	state->eof_reached = false;
+ 	state->markpos_eof = false;
+ 	state->last=state->head;
+ }
+ 
+ /*
+  * tupleconn_performconn: perform connection on tuples
+  * Algorithm: in puttuple has been made list of top parent nodes,
+  * in each iteration we try to find all non-connected tuples which
+  * 'child' attribute is equal to 'parent' attribute in one of parent 
+  * nodes, if so - tuple becomes a child of corresponding parent node.
+  * at end of iteration collected childs becomes the parents for next 
+  * iteration. 
+  * If no childs were find algorithm stops.
+  * Scan for childs in one iteration going on full array of stored 
+  * tuples - this preserves order of tuples from subplan. for example 
+  * if subplan was alphabetically sorted, childs on one level of each 
+  * parent will be also alphabetically sorted.
+  * In case of file storage at end of algorithm all tuples resides only 
+  * on tape.
+  * 
+  * Clause was:
+  * CONNECT BY PRIOR expr Op expr
+  * here:
+  * PRIOR expr is - parentExpr, checked against parent tuple
+  * Op - is connOp operator, performs comparation
+  * expr (the rest) - is childExpr, checked against child tuple
+  */
+ void
+ tupleconn_performconn(Tupleconnstate *state, 
+ 					Node *parentExpr, Node *childExpr, ExprContext *econtext){
+   int ok=0,i,j,ti,level=1;
+ 	ConnectTree **t;
+ 	int32 is_parent;
+ 	Datum dt1,dt2;
+ 	bool p_isnull,isnull,conn;
+ 	HeapTuple pnt,chld;
+ 	TupleDesc tupDesc;
+ 	TupleTableSlot *slot;
+ 	
+ 	if(!state->head) return; /* trivial case, don't connect anything */
+ 
+ 	/* check status */
+ 	switch(state->status){
+ 		case TCS_READMEM:
+ 		case TCS_READFILE:
+ 			break;
+ 		default:
+ 			elog(ERROR,"tupleconn: invalid state in performconn (internal error)");
+ 	}
+ 
+ 	tupDesc=state->tupDesc;
+ 	
+ 	/* make for ExecExvalExpr temporary slot*/
+ 	slot=makeNode(TupleTableSlot);
+ 	slot->ttc_tupleDescriptor = tupDesc;
+ 	slot->ttc_buffer = InvalidBuffer;
+ 	slot->ttc_shouldFreeDesc = false;
+ 	slot->ttc_shouldFree = true;
+ 	slot->ttc_descIsNew = false;
+ 	
+ 	/* set slot for ExecExvalExpr */
+ 	econtext->ecxt_scantuple = slot;
+ 	
+   while(!ok){
+ 	  ok=1;  
+  		for(i=0;i<state->tupcount;i++){//scan through array of tuples
+ 			/* skip already connected and null tuples */
+      	if(!state->used[i]){							
+ 				ResetExprContext(econtext);
+ 				/* get tuple for connecting */
+ 				if(state->status == TCS_READMEM) 
+ 					chld=(HeapTuple)state->memtuples[i];
+ 	  		else chld=READTUP(state,i);
+ 				slot->val = chld;
+ 				/* get value of child expr,
+ 				 * in case of colum name is equal to 
+ 			   * dt1=heap_getattr(chld, attno1, tupDesc, &isnull); 
+ 				 */
+ 				dt1 = ExecEvalExpr(childExpr,econtext,&isnull,NULL);
+ 				conn=false;
+ 				if(!isnull){
+ 					/* scan through nodes array of previous level,
+ 					 * until connect it or array is exhausted
+ 					 */
+ 				  for(j=0;j<state->pntlast && !conn;j++){
+ 						/* get parent tuple */
+ 						if(state->status == TCS_READMEM) 
+ 						  pnt=(HeapTuple)state->memtuples[state->pnt[j]->tuple];
+ 					  else pnt=state->out_store[state->pnt[j]->tuple].tup;
+ 						
+ 						Assert(pnt!=NULL); /*elog(ERROR,"tupleconn: parent tuple is null (internal error)");*/
+ 
+ 						slot->val = pnt;
+ 						/* get value of parent expr */
+ 						dt2 = ExecEvalExpr(parentExpr,econtext,&p_isnull,NULL);
+ 						if(!p_isnull){
+ 							/* apply connOp operator on parent and child expr results,
+ 							 * if 0 (in case of EQ means they're equal) connect them
+ 							 */
+ 							is_parent=DatumGetInt32(FunctionCall2(&state->connFn,dt1,dt2));
+ 							if(is_parent==0){
+ 		        	  ok=0;
+ 								/* stop scan of parents */
+ 								conn=true;
+ 								/* connect tuples (make node of the connect tree)*/
+  			        	state->chld[state->chldlast++]=add_sub(state->pnt[j],i,level+1);
+ 								state->used[i]=1;
+ 								if(state->chldlast==state->chldsize){
+ 									/* grow array of connected tuples as necessary */
+ 									state->chldsize *= 2;
+ 									state->chld = (ConnectTree **)
+ 											repalloc(state->chld,
+ 									 		state->chldsize * sizeof(ConnectTree **));
+ 								}
+   	       		}
+ 						}
+  		    	}
+ 				}else{
+ 					/* mark it as used since it has null child value and can't be 
+ 					 * connected to any node
+ 					 * may be better to add nullcheck on this field at parse stage to whereClause, 
+ 					 */
+ 					state->used[i]=1;
+ 				}
+ 				
+ 				/* this tuple will never become parent for next level (because of NULL child field)
+ 				 * or is wasn't used at current level, 
+ 				 * free it for saving memory in case of operation on tape
+ 				 */
+ 				if(state->status == TCS_READFILE && 
+ 							(state->used[i]!=1 || (state->used[i]==1 && isnull))){
+ 					pfree(chld);					
+ 					state->out_store[i].tup=NULL;
+ 					FREEMEM(state,state->out_store[i].len);
+ 				}
+ 			}
+  	  }
+ 		/* swap pnt & chld arrays */
+ 		t=state->pnt;
+ 		ti=state->pntsize;
+ 
+ 		if(state->status == TCS_READFILE){
+ 			/* remove unneeded parent nodes */
+ 			for(i=0;i<state->pntlast;i++){
+ 				pfree(state->out_store[state->pnt[i]->tuple].tup);
+ 				state->out_store[state->pnt[i]->tuple].tup=NULL;
+ 				FREEMEM(state,state->out_store[state->pnt[i]->tuple].len);
+ 			}
+ 		}
+    	state->pnt=state->chld;
+ 		state->pntsize=state->chldsize;
+ 		state->pntlast=state->chldlast;
+ 
+ 		state->chld=t;
+ 		state->chldsize=ti;
+ 		state->chldlast=0;
+ 
+  	  level++;
+  	}
+ 	/* free anything temporal */
+ 	ResetExprContext(econtext);
+ 	pfree(slot);
+ 	/* find _level_ attribute, will be used at setlevel() below */
+ 	for(i=0;i<state->tupDesc->natts;i++){
+ 		if(!strcmp(state->tupDesc->attrs[i]->attname.data,"_level_")) state->level=i;
+ 	}
+ }
+ 
+ /* set _level_ column to proper value 
+  * first decompose tuple. if we're found _level_ column,
+  * set it to proper value. them form new tuple.
+  * in tape case free original tuple.
+  */
+ void *
+ setlevel(Tupleconnstate *state,int level, void *tuple,bool free_it){
+ #define REALLOCATTRS 64
+ 	Datum valuesArray[REALLOCATTRS];
+ 	char nullsArray[REALLOCATTRS];
+ 	Datum *values;
+ 	char *nulls;
+ 	HeapTuple newtup,tup = (HeapTuple)tuple;
+ 	TupleDesc tdesc = state->tupDesc;
+ 	int natts,i;
+ 
+ 
+ 	if(!tuple) return NULL;
+ 	natts = tdesc->natts;
+ 	/* prepare arrays */
+ 	if(natts>REALLOCATTRS){
+ 		values = palloc(natts * sizeof(Datum));
+ 		nulls = palloc(natts * sizeof(char));
+ 	}else{
+ 		values = valuesArray;
+ 		nulls = nullsArray;
+ 	}
+ 	/* decompose tuple and substitute attr _level_ with real value */
+ 	for (i = 0; i < natts; i++){
+ 		bool		isnull;
+ 		if(i != state->level){
+ 			values[i] = heap_getattr(tup,
+ 								 i + 1,
+ 								 tdesc,
+ 								 &isnull);
+ 			if (isnull)
+ 				nulls[i] = 'n';
+ 			else
+ 				nulls[i] = ' ';
+ 		}else{
+ 			values[state->level]=Int32GetDatum(level);
+ 			nulls[i] = ' ';
+ 		}
+ 	}
+ 	/* form new */
+ 	tup = heap_formtuple(state->tupDesc,values,nulls);
+ 
+ 	if(natts>REALLOCATTRS){
+ 		pfree(values);
+ 		pfree(nulls);
+ 	}
+ 	if(free_it){
+ 		/* make full copy of modified tuple and free original */
+ 		newtup = heap_copytuple(tup);
+ 		tup = newtup;
+ 		FREEMEM(state,((HeapTuple)tuple)->t_len+HEAPTUPLESIZE);
+ 		pfree(tuple);
+ 	}
+ 	
+   return tup;
+ }
+ /*
+  * Fetch the next tuple in forward only direction.
+  * Returns NULL if no more tuples. If should_free is set, the
+  * caller must pfree the returned tuple when done with it.
+  */
+ /* FIXME: add backward direction in future
+  */
+ void *
+ tupleconn_gettuple(Tupleconnstate *state,
+ 					bool *should_free)
+ {
+ 	void	   *tup=NULL;
+ 	int 		level=0;
+ 
+ 	if (state->eof_reached || !state->head) return NULL;
+ 	
+ 	/* check status */
+ 	switch (state->status)
+ 	{
+ 		case TCS_READMEM:
+ 			*should_free = false;
+ 			break;
+ 		case TCS_READFILE:
+ 			*should_free = true;
+ 			break;
+ 		default:
+ 			elog(ERROR, "tupleconn_gettuple: invalid state");
+ 			return NULL;		/* keep compiler happy */
+ 	}
+ 
+ 	while(!tup) {
+ 		if(!state->skip_node) {
+ 			if(state->status == TCS_READMEM){
+ 				tup=state->memtuples[state->last->tuple];
+ 			}else{
+ 				tup=READTUP(state,state->last->tuple);
+ 				/* will be freed in setlevel(), but setlevel()
+ 				 * can't clear tuple's out_store[] cell, so clear it here 
+ 				 */
+ 				state->out_store[state->last->tuple].tup=NULL;
+ 			}
+ 			level=state->last->level;
+ 		}
+ 		if(!state->skip_node && state->last->sub) state->last=state->last->sub;
+ 		else if(state->last->next){
+ 			state->last=state->last->next;
+ 			state->skip_node = false;
+ 		}	else if(state->last->sup){
+ 		  state->last=state->last->sup;
+ 			state->skip_node=true;
+ 		}	else {
+ 		  state->eof_reached = true;
+ 		  break;
+ 		}
+ 	}				
+ 	return setlevel(state,level,tup,*should_free);
+ }
+ 
+ /*
+  * dumptuples - remove tuples from memory and write to tape
+  */
+ static void
+ dumptuples(Tupleconnstate *state)
+ {
+ 	int			i,j;
+ 	bool	b;
+ 	for (i = 0, j = 0; i < state->tupcount; i++){
+ 		/* don't free pnt list, because we will use it soon, in tupleconn_performconn() */
+ 		if(j < state->pntlast && i == state->pnt[j]->tuple){
+ 			b=false;
+ 			j++;
+ 		}else b=true;
+ 		WRITETUP(state, state->memtuples[i],b,i);
+ 	}
+ }
+ 
+ /*
+  * tupleconn_rescan		- rewind and replay the scan
+  */
+ void
+ tupleconn_rescan(Tupleconnstate *state)
+ {
+ 
+ 	/* check status */
+ 	switch (state->status)
+ 	{
+ 		case TCS_READMEM:
+ 		case TCS_READFILE:
+ 			break;
+ 		default:
+ 			elog(ERROR, "tupleconn_rescan: invalid state");
+ 			break;
+ 	}
+ 	state->eof_reached = false;
+ 	state->markpos_eof = false;
+ 	
+ 	state->last = state->head;
+ }
+ 
+ /*
+  * tupleconn_markpos	- saves current position in the tuple sequence
+  */
+ void
+ tupleconn_markpos(Tupleconnstate *state)
+ {
+ 	Assert(state->randomAccess);
+ 
+ 	/* check status */
+ 	switch (state->status)
+ 	{
+ 		case TCS_READMEM:
+ 		case TCS_READFILE:
+ 			break;
+ 		default:
+ 			elog(ERROR, "tupleconn_markpos: invalid state");
+ 			break;
+ 	}
+ 	state->markpos_eof = state->eof_reached;
+ 
+ 	/* file/memtuples positions can be retrieved by state->last
+ 	 * so don't save them
+ 	 */
+ 	state->markpos_last = state->last;
+ 	state->markpos_skip = state->skip_node;
+ }
+ 
+ /*
+  * tupleconn_restorepos - restores current position in connection tree to
+  *						  last saved position
+  */
+ void
+ tupleconn_restorepos(Tupleconnstate *state)
+ {
+ 	Assert(state->randomAccess);
+ 
+ 	/* check status */
+ 	switch (state->status)
+ 	{
+ 		case TCS_READMEM:
+ 		case TCS_READFILE:
+ 			break;
+ 		default:
+ 			elog(ERROR, "tupleconn_restorepos: invalid state");
+ 			break;
+ 	}
+ 	state->eof_reached = state->markpos_eof;
+ 
+ 	state->last = state->markpos_last;
+ 	state->skip_node = state->markpos_skip;
+ }
+ 
+ /*
+  * Routines specialized for HeapTuple case
+  */
+ 
+ static void *
+ copytup_heap(Tupleconnstate *state, void *tup)
+ {
+ 	HeapTuple	tuple = (HeapTuple) tup;
+ 
+ 	USEMEM(state, HEAPTUPLESIZE + tuple->t_len);
+ 	return (void *) heap_copytuple(tuple);
+ }
+ 
+ /*
+  * tree building procedures
+  */
+  
+ /* add new node next to given, set and return it */
+ ConnectTree *
+ add_next(ConnectTree *tr,int tup,int level){
+   ConnectTree *t;
+   
+   t=palloc(sizeof(ConnectTree));
+   memset(t,0,sizeof(ConnectTree));
+   t->tuple=tup;
+   t->level=level;
+   if(tr){
+     tr->next=t;
+     t->prev=tr;
+   }
+   return t;
+ }
+ 
+ /* add new node as child to given, set and return it */
+ ConnectTree *
+ add_sub(ConnectTree *tr,int tup,int level){
+   ConnectTree *t,*t1;
+   
+   t=palloc(sizeof(ConnectTree));
+   memset(t,0,sizeof(ConnectTree));
+   t->tuple=tup;
+   t->level=level;
+   if(!tr->sub){
+     tr->sub=t;  
+     t->sup=tr;
+   }else{
+     for(t1=tr->sub;t1->next;t1=t1->next);
+     t1->next=t;
+     t->prev=t1;
+     t->sup=tr;
+   }
+   return t;
+ }
+ 
+ 
+ /*
+  * File storage procedures
+  * this part taken from tuplestore.c, with addition for 
+  * maintaining out_store[].
+  * because of all on-tape tuple placement info is stored in 
+  * out_store[], don't bother to write it on a tape, only
+  * tuple body is stored.
+  */
+  
+ /*
+  * We don't bother to write the HeapTupleData part of the tuple.
+  */
+ 
+ static void
+ writetup_heap(Tupleconnstate *state, void *tup, bool free, int idx)
+ {
+ 	HeapTuple	tuple = (HeapTuple) tup;
+ 
+ 	/* fill placement info */
+ 	state->out_store[idx].len = tuple->t_len;
+ 	BufFileTell(state->myfile,
+ 				&state->out_store[idx].fileno,
+ 				&state->out_store[idx].offs);
+ 	
+ 	if (BufFileWrite(state->myfile, (void *) tuple->t_data,
+ 					 tuple->t_len) != (size_t) tuple->t_len)
+ 		elog(ERROR, "tupleconn: write failed");
+ 	
+ 	/* explanation in dumptuples */
+ 	if(free){
+ 		FREEMEM(state, HEAPTUPLESIZE + tuple->t_len);
+ 		heap_freetuple(tuple);
+ 		state->out_store[idx].tup = NULL;		
+ 	}else{
+ 		state->out_store[idx].tup = (HeapTuple)tup;
+ 	}
+ 	
+ }
+ 
+ static void *
+ readtup_heap(Tupleconnstate *state, int idx)
+ {
+ 	unsigned int tuplen = state->out_store[idx].len + HEAPTUPLESIZE;
+ 	/* add block readings */
+ 	HeapTuple	tuple = (HeapTuple) palloc(tuplen);
+ 
+ 	USEMEM(state, tuplen);
+ 	/* reconstruct the HeapTupleData portion */
+ 	tuple->t_len = state->out_store[idx].len ;
+ 	ItemPointerSetInvalid(&(tuple->t_self));
+ 	tuple->t_datamcxt = CurrentMemoryContext;
+ 	tuple->t_data = (HeapTupleHeader) (((char *) tuple) + HEAPTUPLESIZE);
+ 	/* seek to the tuple */
+ 	if(BufFileSeek(state->myfile,
+ 			state->out_store[idx].fileno,
+ 			state->out_store[idx].offs,SEEK_SET)!=0)
+ 		elog(ERROR,"tupleconn: can't seek in readtup_heap");
+ 
+ 	/* read in the tuple proper */
+ 	if (BufFileRead(state->myfile, (void *) tuple->t_data,
+ 					tuple->t_len) != (size_t) tuple->t_len)
+ 		elog(ERROR, "tupleconn: unexpected end of data");
+ 
+ 	state->out_store[idx].tup = tuple;
+ 	return (void *) tuple;
+ }
+ 
+ /*
+  * Select comparation function
+  * made from SelectSortFunction() in nodeSort.c.
+  * differences is the name:) and strategy of operator
+  * being searched (here is BTEqualStrategyNumber)
+  */
+  
+ void
+ SelectConnFunction(Oid connOperator,
+ 				   RegProcedure *connFunction,
+ 				   SortFunctionKind *kind)
+ {
+ 	Relation	relation;
+ 	HeapScanDesc scan;
+ 	ScanKeyData skey[1];
+ 	HeapTuple	tuple;
+ 	Form_pg_operator optup;
+ 	Oid			opclass = InvalidOid;
+ 
+ 	ScanKeyEntryInitialize(&skey[0], 0x0,
+ 						   Anum_pg_amop_amopopr,
+ 						   F_OIDEQ,
+ 						   ObjectIdGetDatum(connOperator));
+ 
+ 	relation = heap_openr(AccessMethodOperatorRelationName, AccessShareLock);
+ 	scan = heap_beginscan(relation, SnapshotNow, 1, skey);
+ 
+ 	while ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
+ 	{
+ 		Form_pg_amop aform = (Form_pg_amop) GETSTRUCT(tuple);
+ 
+ 		if (!opclass_is_btree(aform->amopclaid))
+ 			continue;
+ 		if (aform->amopstrategy == BTEqualStrategyNumber)
+ 		{
+ 			opclass = aform->amopclaid;
+ 			*kind = SORTFUNC_CMP;
+ 			break;				/* done looking */
+ 		}
+ 	}
+ 
+ 	heap_endscan(scan);
+ 	heap_close(relation, AccessShareLock);
+ 
+ 	if (OidIsValid(opclass))
+ 	{
+ 		/* Found a suitable opclass, get its comparator support function */
+ 		tuple = SearchSysCache(AMPROCNUM,
+ 							   ObjectIdGetDatum(opclass),
+ 							   Int16GetDatum(BTORDER_PROC),
+ 							   0, 0);
+ 		if (HeapTupleIsValid(tuple))
+ 		{
+ 			Form_pg_amproc aform = (Form_pg_amproc) GETSTRUCT(tuple);
+ 
+ 			*connFunction = aform->amproc;
+ 			ReleaseSysCache(tuple);
+ 			Assert(RegProcedureIsValid(*connFunction));
+ 			return;
+ 		}
+ 	}
+ 
+ 	/*
+ 	 * Can't find a comparator, so use the operator as-is.  Decide whether
+ 	 * it is forward or reverse conn by looking at its name (grotty, but
+ 	 * this only matters for deciding which end NULLs should get conned
+ 	 * to).
+ 	 */
+ 	tuple = SearchSysCache(OPEROID,
+ 						   ObjectIdGetDatum(connOperator),
+ 						   0, 0, 0);
+ 	if (!HeapTupleIsValid(tuple))
+ 		elog(ERROR, "SelectConnFunction: cache lookup failed for operator %u",
+ 			 connOperator);
+ 	optup = (Form_pg_operator) GETSTRUCT(tuple);
+ 	if (strcmp(NameStr(optup->oprname), ">") == 0)
+ 		*kind = SORTFUNC_REVLT;
+ 	else
+ 		*kind = SORTFUNC_LT;
+ 	*connFunction = optup->oprcode;
+ 	ReleaseSysCache(tuple);
+ 
+ 	Assert(RegProcedureIsValid(*connFunction));
+ }
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/executor/nodeConn.h postgresql-7.3rc1-evg/src/include/executor/nodeConn.h
*** postgresql-7.3rc1/src/include/executor/nodeConn.h	Thu Jan  1 04:00:00 1970
--- postgresql-7.3rc1-evg/src/include/executor/nodeConn.h	Thu Nov 28 19:15:21 2002
***************
*** 0 ****
--- 1,27 ----
+ /*-------------------------------------------------------------------------
+  *
+  * nodeSort.h
+  *
+  *
+  *
+  * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+  * Portions Copyright (c) 1994, Regents of the University of California
+  *
+  * $Id: nodeSort.h,v 1.14 2001/11/05 17:46:33 momjian Exp $
+  *
+  *-------------------------------------------------------------------------
+  */
+ #ifndef NODECONN_H
+ #define NODECONN_H
+ 
+ #include "nodes/plannodes.h"
+ 
+ extern TupleTableSlot *ExecConn(Conn *node);
+ extern bool ExecInitConn(Conn *node, EState *estate, Plan *parent);
+ extern int	ExecCountSlotsConn(Conn *node);
+ extern void ExecEndConn(Conn *node);
+ extern void ExecConnMarkPos(Conn *node);
+ extern void ExecConnRestrPos(Conn *node);
+ extern void ExecReScanConn(Conn *node, ExprContext *exprCtxt, Plan *parent);
+ 
+ #endif   /* NODECONN_H */
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/nodes/execnodes.h postgresql-7.3rc1-evg/src/include/nodes/execnodes.h
*** postgresql-7.3rc1/src/include/nodes/execnodes.h	Thu Sep  5 01:31:42 2002
--- postgresql-7.3rc1-evg/src/include/nodes/execnodes.h	Thu Nov 28 19:15:21 2002
***************
*** 709,714 ****
--- 709,728 ----
  } SortState;
  
  /* ----------------
+  *	 ConnectState information
+  *
+  *		conn_Done		indicates whether connection has been performed yet
+  *		tupleconnstate	private state of tuplesort.c
+  * ----------------
+  */
+ typedef struct ConnectState
+ {
+  	CommonScanState csstate;	/* its first field is NodeTag */
+  	bool		conn_Done;
+  	void	   *tupleconnstate;
+ } ConnectState;
+  
+ /* ----------------
   *	 UniqueState information
   *
   *		Unique nodes are used "on top of" sort nodes to discard
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/nodes/nodes.h postgresql-7.3rc1-evg/src/include/nodes/nodes.h
*** postgresql-7.3rc1/src/include/nodes/nodes.h	Fri Oct 11 09:16:44 2002
--- postgresql-7.3rc1-evg/src/include/nodes/nodes.h	Thu Nov 28 19:15:21 2002
***************
*** 41,46 ****
--- 41,47 ----
  	T_Limit,
  	T_Material,
  	T_Sort,
+ 	T_Conn,
  	T_Agg,
  	T_Unique,
  	T_Hash,
***************
*** 58,63 ****
--- 59,65 ----
  	T_Fjoin,
  	T_Expr,
  	T_Var,
+ 	T_FakeVar,
  	T_Oper,
  	T_Const,
  	T_Param,
***************
*** 111,116 ****
--- 113,119 ----
  	T_AggState,
  	T_GroupState,
  	T_SortState,
+ 	T_ConnectState,
  	T_UniqueState,
  	T_HashState,
  	T_TidScanState,
***************
*** 239,245 ****
  	T_InsertDefault,
  	T_CreateOpClassItem,
  	T_CompositeTypeStmt,
! 
  	/*
  	 * TAGS FOR FUNCTION-CALL CONTEXT AND RESULTINFO NODES (see fmgr.h)
  	 */
--- 242,248 ----
  	T_InsertDefault,
  	T_CreateOpClassItem,
  	T_CompositeTypeStmt,
! 	T_HierClause,
  	/*
  	 * TAGS FOR FUNCTION-CALL CONTEXT AND RESULTINFO NODES (see fmgr.h)
  	 */
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/nodes/parsenodes.h postgresql-7.3rc1-evg/src/include/nodes/parsenodes.h
*** postgresql-7.3rc1/src/include/nodes/parsenodes.h	Tue Oct 15 03:14:35 2002
--- postgresql-7.3rc1-evg/src/include/nodes/parsenodes.h	Thu Nov 28 19:15:21 2002
***************
*** 85,90 ****
--- 85,91 ----
  	Node	   *setOperations;	/* set-operation tree if this is top level
  								 * of a UNION/INTERSECT/EXCEPT query */
  
+  	Node     *hierClause; /* CONNECT BY/START WITH clause */
  	/*
  	 * If the resultRelation turns out to be the parent of an inheritance
  	 * tree, the planner will add all the child tables to the rtable and
***************
*** 612,618 ****
   */
  typedef SortClause GroupClause;
  
! 
  /*****************************************************************************
   *		Optimizable Statements
   *****************************************************************************/
--- 613,635 ----
   */
  typedef SortClause GroupClause;
  
! /*
!  * HierClause -
!  *	   representation of CONNECT BY/START WITH clauses
!  * parsed:
!  * CONNECT BY ( PRIOR expr ) ( op ) ( expr )
!  * first () is parentExpr
!  * second () is connOpName
!  * third () is childExpr
!  */
! typedef struct HierClause
! {
!  	NodeTag type;
! 	Node *parentExpr,*childExpr;
! 	Node *connOpName;
!  	Node	*startQual;		/* Quals for heads */
!  	Oid   connOp;	/* operator for comparing parent & child exprs*/
! } HierClause;
  /*****************************************************************************
   *		Optimizable Statements
   *****************************************************************************/
***************
*** 698,703 ****
--- 715,721 ----
  	Node	   *whereClause;	/* WHERE qualification */
  	List	   *groupClause;	/* GROUP BY clauses */
  	Node	   *havingClause;	/* HAVING conditional-expression */
+  	List		 *hierClause;    /* CONNECT BY , START WITH clauses*/
  
  	/*
  	 * These fields are used in both "leaf" SelectStmts and upper-level
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/nodes/plannodes.h postgresql-7.3rc1-evg/src/include/nodes/plannodes.h
*** postgresql-7.3rc1/src/include/nodes/plannodes.h	Thu Sep  5 01:31:44 2002
--- postgresql-7.3rc1-evg/src/include/nodes/plannodes.h	Thu Nov 28 19:15:21 2002
***************
*** 50,55 ****
--- 50,56 ----
   *		SetOp					SetOpState				setopstate;
   *		Limit					LimitState				limitstate;
   *		Hash					HashState				hashstate;
+  *		Conn					ConnectState		connstate;
   *
   * ----------------------------------------------------------------
   */
***************
*** 365,370 ****
--- 366,385 ----
  	int			keycount;
  	SortState  *sortstate;
  } Sort;
+ 
+ /* ----------------
+  *		conn node
+  * ----------------
+  */
+ typedef struct Conn
+ {
+  	Plan		plan;
+  	ConnectState *connstate;
+  	List	*startQual;   /* qual conditions for heads */
+ 	Node	*parentExpr;
+ 	Node	*childExpr;
+  	Oid		connOp;
+ }Conn;
  
  /* ----------------
   *		unique node
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/nodes/primnodes.h postgresql-7.3rc1-evg/src/include/nodes/primnodes.h
*** postgresql-7.3rc1/src/include/nodes/primnodes.h	Thu Sep 19 02:35:24 2002
--- postgresql-7.3rc1-evg/src/include/nodes/primnodes.h	Thu Nov 28 19:15:21 2002
***************
*** 261,266 ****
--- 261,291 ----
  	AttrNumber	varoattno;		/* original value of varattno */
  } Var;
  
+ /* FakeVar is same as Var, with several exeptions:
+  *
+  * it's not fetched from the tables.
+  * it's not passed to scan nodes so them don't see it.
+  *		 when we comes to moment when we must fetch value from nowhere 
+  * 		(because of nonexistence of FakeVar as relation attr), we just return
+  * 		0 as Datum. this made at executor/execQual.c:ExecEvalFakeVar().
+  * FakeVar exist only for checking attrs in scan tuple with baseing on targetlist,
+  * 		not in outer or inner relations. so there is no varno.
+  * it never rewritten because of it's based on targetlist where it's present.
+  * 		so there is no *old fields.
+  * now supproted only INT4 type FakeVars (see at executor/execQual.c:ExecEvalFakeVar()).
+  * 		but may be extended.
+  */
+ typedef struct FakeVar
+ {
+ 	NodeTag		type;
+ 	AttrNumber	varattno;		/* attribute number of this var, or zero
+ 								 * for all */
+ 	Oid			vartype;		/* pg_type tuple OID for the type of this
+ 								 * var */
+ 	int32		vartypmod;		/* pg_attribute typmod value */
+ 
+ } FakeVar;
+ 
  /*
   * Const
   */
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/optimizer/clauses.h postgresql-7.3rc1-evg/src/include/optimizer/clauses.h
*** postgresql-7.3rc1/src/include/optimizer/clauses.h	Wed Sep 11 19:48:55 2002
--- postgresql-7.3rc1-evg/src/include/optimizer/clauses.h	Thu Nov 28 19:15:21 2002
***************
*** 42,47 ****
--- 42,49 ----
  extern bool contain_agg_clause(Node *clause);
  extern List *pull_agg_clause(Node *clause);
  
+ extern bool contain_fake_column(Node *clause);
+  
  extern bool expression_returns_set(Node *clause);
  
  extern bool contain_subplans(Node *clause);
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/optimizer/cost.h postgresql-7.3rc1-evg/src/include/optimizer/cost.h
*** postgresql-7.3rc1/src/include/optimizer/cost.h	Thu Sep  5 01:31:44 2002
--- postgresql-7.3rc1-evg/src/include/optimizer/cost.h	Thu Nov 28 19:15:21 2002
***************
*** 57,62 ****
--- 57,64 ----
  				  RelOptInfo *baserel);
  extern void cost_sort(Path *path, Query *root,
  		  List *pathkeys, double tuples, int width);
+ extern void cost_conn(Path *path, Query *root,
+  		  List *pathkeys, double tuples, int width);
  extern void cost_nestloop(Path *path, Query *root,
  			  Path *outer_path, Path *inner_path,
  			  List *restrictlist);
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/optimizer/planmain.h postgresql-7.3rc1-evg/src/include/optimizer/planmain.h
*** postgresql-7.3rc1/src/include/optimizer/planmain.h	Thu Sep  5 01:31:45 2002
--- postgresql-7.3rc1-evg/src/include/optimizer/planmain.h	Thu Nov 28 19:15:21 2002
***************
*** 33,38 ****
--- 33,39 ----
  		  Plan *lefttree, int keycount);
  extern Sort *make_sort_from_pathkeys(Query *root, List *tlist,
  						Plan *lefttree, List *pathkeys);
+ extern Conn *make_conn(Query *root, List *tlist, Plan *lefttree);
  extern Agg *make_agg(List *tlist, List *qual, Plan *lefttree);
  extern Group *make_group(List *tlist, bool tuplePerGroup, int ngrp,
  		   AttrNumber *grpColIdx, Plan *lefttree);
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/optimizer/planner.h postgresql-7.3rc1-evg/src/include/optimizer/planner.h
*** postgresql-7.3rc1/src/include/optimizer/planner.h	Fri Jun 21 01:29:51 2002
--- postgresql-7.3rc1-evg/src/include/optimizer/planner.h	Thu Nov 28 19:15:21 2002
***************
*** 23,27 ****
--- 23,29 ----
  
  extern Plan *make_sortplan(Query *parse, List *tlist,
  			  Plan *plannode, List *sortcls);
+ extern Plan *make_connplan(Query *parse, List *tlist,
+  			  Plan *plannode);
  
  #endif   /* PLANNER_H */
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/parser/parse_clause.h postgresql-7.3rc1-evg/src/include/parser/parse_clause.h
*** postgresql-7.3rc1/src/include/parser/parse_clause.h	Fri Jun 21 01:29:51 2002
--- postgresql-7.3rc1-evg/src/include/parser/parse_clause.h	Thu Nov 28 19:15:21 2002
***************
*** 23,28 ****
--- 23,29 ----
  extern Node *transformWhereClause(ParseState *pstate, Node *where);
  extern List *transformGroupClause(ParseState *pstate, List *grouplist,
  					 List *targetlist);
+ extern Node *transformHierClause(ParseState *pstate, Query *qry, List *hier);
  extern List *transformSortClause(ParseState *pstate, List *orderlist,
  					List *targetlist);
  extern List *transformDistinctClause(ParseState *pstate, List *distinctlist,
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/parser/parse_node.h postgresql-7.3rc1-evg/src/include/parser/parse_node.h
*** postgresql-7.3rc1/src/include/parser/parse_node.h	Thu Sep 19 02:35:24 2002
--- postgresql-7.3rc1-evg/src/include/parser/parse_node.h	Thu Nov 28 19:15:21 2002
***************
*** 46,51 ****
--- 46,64 ----
  	bool		p_hasSubLinks;
  	bool		p_is_insert;
  	bool		p_is_update;
+ 
+ 	AttrNumber p_hierLevelColIdx; /* _level_ column number for hier clause, if any. 
+ 																 * set in transformHierClause() */
+ 	bool		p_addTle; /* flag used in parse_expr.c:transformIndirection().
+ 										 * if set all vars going thought will be checked against
+ 										 * current targetlist (field below), and if not present,
+ 										 * added to it as junk.
+ 										 * this is needed in HierClause case. in parent/child Expr and startQual
+ 										 * may be preset attributes which is not in targetlist given by user, so
+ 										 * at transformation stage we'll add them.
+ 										 */
+ 	List		 *p_curTlist; /* current target list (see above)*/
+ 
  	Relation	p_target_relation;
  	RangeTblEntry *p_target_rangetblentry;
  } ParseState;
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/include/utils/tupleconn.h postgresql-7.3rc1-evg/src/include/utils/tupleconn.h
*** postgresql-7.3rc1/src/include/utils/tupleconn.h	Thu Jan  1 04:00:00 1970
--- postgresql-7.3rc1-evg/src/include/utils/tupleconn.h	Thu Nov 28 19:15:21 2002
***************
*** 0 ****
--- 1,64 ----
+ /*-------------------------------------------------------------------------
+  *
+  * tuplestore.h
+  *	  Generalized routines for temporary tuple storage.
+  *
+  * This module handles temporary storage of tuples for purposes such
+  * as Materialize nodes, hashjoin batch files, etc.  It is essentially
+  * a dumbed-down version of tuplesort.c; it does no sorting of tuples
+  * but can only store a sequence of tuples and regurgitate it later.
+  * A temporary file is used to handle the data if it exceeds the
+  * space limit specified by the caller.
+  *
+  * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+  * Portions Copyright (c) 1994, Regents of the University of California
+  *
+  * $Id: tuplestore.h,v 1.6 2001/11/05 17:46:36 momjian Exp $
+  *
+  *-------------------------------------------------------------------------
+  */
+ #ifndef TUPLECONN_H
+ #define TUPLECONN_H
+ 
+ #include "access/htup.h"
+ #include "nodes/nodes.h"
+ #include "executor/executor.h"
+ 
+ /* Tuplestorestate is an opaque type whose details are not known outside
+  * tuplestore.c.
+  */
+ typedef struct Tupleconnstate Tupleconnstate;
+ 
+ /*
+  * Currently we only need to store HeapTuples, but it would be easy
+  * to support the same behavior for IndexTuples and/or bare Datums.
+  */
+ 
+ extern Tupleconnstate *tupleconn_begin_heap(
+ 					  int maxKBytes, TupleDesc tupDesc, Oid connOp);
+ 
+ extern void tupleconn_puttuple(Tupleconnstate *state, void *tuple,int head);
+ 
+ extern void tupleconn_donestoring(Tupleconnstate *state);
+ 
+ extern void *tupleconn_gettuple(Tupleconnstate *state,
+ 					bool *should_free);
+ 
+ #define tupleconn_getheaptuple(state, should_free) \
+ 	((HeapTuple) tupleconn_gettuple(state, should_free))
+ 
+ extern void tupleconn_end(Tupleconnstate *state);
+ 
+ extern void tupleconn_performconn(Tupleconnstate *state, Node *parentExpr, Node *childExpr, ExprContext *econtext);
+ 
+ /*
+  * These routines may only be called if randomAccess was specified 'true'.
+  * Likewise, backwards scan in gettuple/getdatum is only allowed if
+  * randomAccess was specified.
+  */
+ 
+ extern void tupleconn_rescan(Tupleconnstate *state);
+ extern void tupleconn_markpos(Tupleconnstate *state);
+ extern void tupleconn_restorepos(Tupleconnstate *state);
+ 
+ #endif   /* TUPLECONN_H */
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/test/regress/expected/select_hier.out postgresql-7.3rc1-evg/src/test/regress/expected/select_hier.out
*** postgresql-7.3rc1/src/test/regress/expected/select_hier.out	Thu Jan  1 04:00:00 1970
--- postgresql-7.3rc1-evg/src/test/regress/expected/select_hier.out	Thu Nov 28 20:22:52 2002
***************
*** 0 ****
--- 1,45 ----
+ --- create table
+ CREATE TABLE hier (id int4, pnt int4, name varchar(20));
+ --- stuff it
+ INSERT INTO hier VALUES (0, 0, 'root        ');
+ INSERT INTO hier VALUES (1, 0, '(1 leaf l1) ');
+ INSERT INTO hier VALUES (2, 0, '(2 leaf l1) ');
+ INSERT INTO hier VALUES (3, 0, '(3 leaf l1) ');
+ INSERT INTO hier VALUES (4, 1, '(11 leaf l2)');
+ INSERT INTO hier VALUES (5, 1, '(12 leaf l2)');
+ INSERT INTO hier VALUES (6, 1, '(13 leaf l2)');
+ INSERT INTO hier VALUES (7, 6, '(131 leaf l2)');
+ INSERT INTO hier VALUES (8, 6, '(132 leaf l2)');
+ INSERT INTO hier VALUES (9, 5, '(121 leaf l3)');
+ INSERT INTO hier VALUES (10, 5, '(122 leaf l3)');
+ INSERT INTO hier VALUES (11, 5, '(123 leaf l3)');
+ --- hierarchical select 
+ SELECT * FROM hier CONNECT BY PRIOR id = pnt START WITH id = 0;
+  id | pnt |     name      | _level_ 
+ ----+-----+---------------+---------
+   0 |   0 | root          |       1
+   1 |   0 | (1 leaf l1)   |       2
+   4 |   1 | (11 leaf l2)  |       3
+   5 |   1 | (12 leaf l2)  |       3
+   9 |   5 | (121 leaf l3) |       4
+  10 |   5 | (122 leaf l3) |       4
+  11 |   5 | (123 leaf l3) |       4
+   6 |   1 | (13 leaf l2)  |       3
+   7 |   6 | (131 leaf l2) |       4
+   8 |   6 | (132 leaf l2) |       4
+   2 |   0 | (2 leaf l1)   |       2
+   3 |   0 | (3 leaf l1)   |       2
+ (12 rows)
+ 
+ --- hierarchical select with having clause
+ SELECT * FROM hier CONNECT BY PRIOR id = pnt START WITH id = 0 HAVING _level_ <= 2;
+  id | pnt |     name     | _level_ 
+ ----+-----+--------------+---------
+   0 |   0 | root         |       1
+   1 |   0 | (1 leaf l1)  |       2
+   2 |   0 | (2 leaf l1)  |       2
+   3 |   0 | (3 leaf l1)  |       2
+ (4 rows)
+ 
+ --- cleanup
+ DROP TABLE hier;
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/test/regress/parallel_schedule postgresql-7.3rc1-evg/src/test/regress/parallel_schedule
*** postgresql-7.3rc1/src/test/regress/parallel_schedule	Wed Oct 23 01:20:10 2002
--- postgresql-7.3rc1-evg/src/test/regress/parallel_schedule	Thu Nov 28 19:53:24 2002
***************
*** 60,66 ****
  # ----------
  # The fourth group of parallel test
  # ----------
! test: select_into select_distinct select_distinct_on select_implicit select_having subselect union case join aggregates transactions random portals arrays btree_index hash_index
  
  test: privileges
  test: misc
--- 60,66 ----
  # ----------
  # The fourth group of parallel test
  # ----------
! test: select_into select_distinct select_distinct_on select_implicit select_having subselect union case join aggregates transactions random portals arrays btree_index hash_index select_hier
  
  test: privileges
  test: misc
diff -P -r -d -c --exclude-from=exclude postgresql-7.3rc1/src/test/regress/sql/select_hier.sql postgresql-7.3rc1-evg/src/test/regress/sql/select_hier.sql
*** postgresql-7.3rc1/src/test/regress/sql/select_hier.sql	Thu Jan  1 04:00:00 1970
--- postgresql-7.3rc1-evg/src/test/regress/sql/select_hier.sql	Thu Nov 28 20:21:19 2002
***************
*** 0 ****
--- 1,26 ----
+ 
+ --- create table
+ CREATE TABLE hier (id int4, pnt int4, name varchar(20));
+ 
+ --- stuff it
+ INSERT INTO hier VALUES (0, 0, 'root        ');
+ INSERT INTO hier VALUES (1, 0, '(1 leaf l1) ');
+ INSERT INTO hier VALUES (2, 0, '(2 leaf l1) ');
+ INSERT INTO hier VALUES (3, 0, '(3 leaf l1) ');
+ INSERT INTO hier VALUES (4, 1, '(11 leaf l2)');
+ INSERT INTO hier VALUES (5, 1, '(12 leaf l2)');
+ INSERT INTO hier VALUES (6, 1, '(13 leaf l2)');
+ INSERT INTO hier VALUES (7, 6, '(131 leaf l2)');
+ INSERT INTO hier VALUES (8, 6, '(132 leaf l2)');
+ INSERT INTO hier VALUES (9, 5, '(121 leaf l3)');
+ INSERT INTO hier VALUES (10, 5, '(122 leaf l3)');
+ INSERT INTO hier VALUES (11, 5, '(123 leaf l3)');
+ 
+ --- hierarchical select 
+ SELECT * FROM hier CONNECT BY PRIOR id = pnt START WITH id = 0;
+ 
+ --- hierarchical select with having clause
+ SELECT * FROM hier CONNECT BY PRIOR id = pnt START WITH id = 0 HAVING _level_ <= 2;
+ 
+ --- cleanup
+ DROP TABLE hier;
